from typing import Any, Dict, List, Optional, cast, Tuple
import flask
from uuid import uuid4

import dash.development.base_component as bc
import dash_bootstrap_components as dbc
import dash_mantine_components as dmc
from dash import Input, Output, State, callback, ctx, html, no_update, ALL, dcc

import mitzu.model as M
import mitzu.webapp.dependencies as DEPS
import mitzu.webapp.pages.paths as P
import mitzu.webapp.configs as CO
from mitzu.helper import value_to_label
from mitzu.webapp.auth.decorator import restricted
from mitzu.webapp.helper import MITZU_LOCATION, create_form_property_input
import traceback

EXTRA_PROPERTY_CONTAINER = "extra_property_container"
CONNECTION_DELETE_BUTTON = "connection_delete_button"
CONNECTION_TEST_BUTTON = "connection_test_button"
INDEX_TYPE = "connection_property"
TEST_CONNECTION_RESULT = "test_connection_result"
DELETE_CONNECTION_RESULT = "delete_connection_result"
TEST_CONNECTION_LOADING = "test_connection_loading"

CONFIRM_DIALOG_INDEX = "connection_confirm"
CONFIRM_DIALOG_CLOSE = "conn_confirm_dialog_close"
CONFIRM_DIALOG_ACCEPT = "conn_confirm_dialog_accept"


PROP_CONNECTION_ID = "connection_id"
PROP_CONNECTION_NAME = "connection_name"
PROP_CONNECTION_TYPE = "connection_type"
PROP_CATALOG = "catalog"
PROP_REGION = "region"
PROP_S3_STAGING_DIR = "s3_staging_dir"
PROP_WAREHOUSE = "warehouse"
PROP_ROLE = "role"
PROP_HTTP_PATH = "http_path"
PROP_PORT = "port"
PROP_HOST = "host"
PROP_USERNAME = "username"
PROP_PASSWORD = "password"
NOT_REQUIRED_PROPERTIES = [PROP_CATALOG, PROP_PORT, PROP_PASSWORD, PROP_USERNAME]
CON_TYPE_BLACKLIST = [M.ConnectionType.FILE, M.ConnectionType.SQLITE]


MIN_LENGTH = 4
MAX_LENGTH = 100


def create_url_param(values: Dict[str, Any], property: str) -> str:
    return f"{property}={values[property]}"


def create_connection_from_values(values: Dict[str, Any]) -> M.Connection:
    con_type = M.ConnectionType.parse(values[PROP_CONNECTION_TYPE])
    url_params = ""
    extra_configs = {}
    if con_type == M.ConnectionType.ATHENA:
        url_params = "&".join(
            [
                create_url_param(values, PROP_S3_STAGING_DIR),
                create_url_param(values, PROP_REGION),
            ]
        )
    elif con_type == M.ConnectionType.DATABRICKS:
        extra_configs = {PROP_HTTP_PATH: values[PROP_HTTP_PATH]}
    elif con_type == M.ConnectionType.SNOWFLAKE:
        url_params = "&".join(
            [
                create_url_param(values, PROP_WAREHOUSE),
                create_url_param(values, PROP_ROLE),
            ]
        )
    return M.Connection(
        connection_name=values[PROP_CONNECTION_NAME],
        connection_type=con_type,
        host=values[PROP_HOST],
        port=values[PROP_PORT],
        id=values[PROP_CONNECTION_ID],
        catalog=values[PROP_CATALOG],
        user_name=values[PROP_USERNAME],
        secret_resolver=M.ConstSecretResolver(values[PROP_PASSWORD]),
        url_params=url_params,
        extra_configs=extra_configs,
    )


def create_connection_extra_inputs(
    con_type: Optional[M.ConnectionType], con: Optional[M.Connection]
) -> List[bc.Component]:
    if con_type is None:
        return []
    if con_type == M.ConnectionType.ATHENA:
        return [
            create_form_property_input(
                index_type=INDEX_TYPE,
                property=PROP_REGION,
                type="text",
                required=True,
                icon_cls="bi bi-globe",
                value=con.get_url_param(PROP_REGION) if con is not None else None,
                minlength=4,
            ),
            create_form_property_input(
                index_type=INDEX_TYPE,
                property=PROP_S3_STAGING_DIR,
                type="text",
                required=True,
                icon_cls="bi bi-bucket",
                value=(
                    con.get_url_param(PROP_S3_STAGING_DIR) if con is not None else None
                ),
                minlength=4,
            ),
        ]
    if con_type == M.ConnectionType.DATABRICKS:
        return [
            create_form_property_input(
                index_type=INDEX_TYPE,
                property=PROP_HTTP_PATH,
                type="text",
                required=True,
                icon_cls="bi bi-link",
                value=con.extra_configs.get(PROP_HTTP_PATH)
                if con is not None
                else None,
                minlength=4,
            )
        ]
    if con_type == M.ConnectionType.SNOWFLAKE:
        return [
            create_form_property_input(
                index_type=INDEX_TYPE,
                property=PROP_WAREHOUSE,
                type="text",
                required=True,
                icon_cls="bi bi-house",
                value=con.get_url_param(PROP_WAREHOUSE) if con is not None else None,
                minlength=4,
            ),
            create_form_property_input(
                index_type=INDEX_TYPE,
                property=PROP_ROLE,
                type="text",
                required=True,
                icon_cls="bi bi-file-earmark-person-fill",
                value=con.get_url_param(PROP_ROLE) if con is not None else None,
                minlength=4,
            ),
        ]

    return []


def create_delete_button(connection: Optional[M.Connection]) -> bc.Component:
    if connection is not None:
        return dbc.Button(
            [html.B(className="bi bi-x-circle"), " Delete connection"],
            id=CONNECTION_DELETE_BUTTON,
            color="danger",
        )
    else:
        return html.Div()


def create_confirm_dialog():
    return dbc.Modal(
        [
            dbc.ModalBody(
                "Do you really want to delete the connection?", class_name="lead"
            ),
            dbc.ModalFooter(
                [
                    dbc.Button(
                        "Close",
                        id=CONFIRM_DIALOG_CLOSE,
                        size="sm",
                        color="secondary",
                        class_name="me-1",
                    ),
                    dbc.Button(
                        "Delete",
                        id=CONFIRM_DIALOG_ACCEPT,
                        size="sm",
                        color="danger",
                        href=P.CONNECTIONS_PATH,
                        external_link=True,
                    ),
                ]
            ),
        ],
        id=CONFIRM_DIALOG_INDEX,
        is_open=False,
    )


def create_manage_connection_component(
    con: Optional[M.Connection],
) -> bc.Component:
    con_type_opts = [
        {"label": ct.name.title(), "value": ct.name}
        for ct in M.ConnectionType
        if ct not in CON_TYPE_BLACKLIST
    ]
    if CO.SETUP_SAMPLE_PROJECT:
        con_type_opts.append(
            {
                "label": M.ConnectionType.SQLITE.name.title(),
                "value": M.ConnectionType.SQLITE.name,
            }
        )
    def_con_type = M.ConnectionType.SNOWFLAKE
    return html.Div(
        [
            html.Div(
                create_form_property_input(
                    index_type=INDEX_TYPE,
                    property=PROP_CONNECTION_ID,
                    icon_cls="bi bi-info-circle",
                    type="text",
                    value=con.id if con is not None else str(uuid4())[-12:],
                    disabled=True,
                    class_name="d-none",
                ),
                className="d-none",
            ),
            create_form_property_input(
                index_type=INDEX_TYPE,
                property=PROP_CONNECTION_NAME,
                icon_cls="bi bi-card-text",
                type="text",
                required=True,
                value=con.connection_name if con is not None else None,
                minlength=4,
            ),
            create_form_property_input(
                index_type=INDEX_TYPE,
                property=PROP_CONNECTION_TYPE,
                icon_cls="bi bi-gear-wide",
                component_type=dmc.Select,
                required=True,
                value=(
                    con.connection_type.name if con is not None else def_con_type.name
                ),
                searchable=True,
                data=con_type_opts,
                size="xs",
            ),
            create_form_property_input(
                index_type=INDEX_TYPE,
                property=PROP_HOST,
                icon_cls="bi bi-link",
                type="text",
                required=True,
                value=con.host if con is not None else None,
                minlength=4,
            ),
            create_form_property_input(
                index_type=INDEX_TYPE,
                property=PROP_PORT,
                icon_cls="bi bi-hash",
                type="number",
                value=con.port if con is not None else None,
            ),
            create_form_property_input(
                index_type=INDEX_TYPE,
                property=PROP_CATALOG,
                icon_cls="bi bi-journals",
                type="text",
                value=con.catalog if con is not None else None,
            ),
            html.Hr(),
            create_form_property_input(
                index_type=INDEX_TYPE,
                property=PROP_USERNAME,
                icon_cls="bi bi-person-fill",
                type="text",
                value=con.user_name if con is not None else None,
                minlength=4,
            ),
            create_form_property_input(
                index_type=INDEX_TYPE,
                property=PROP_PASSWORD,
                icon_cls="bi bi-lock-fill",
                type="password",
                value=con.password if con is not None else None,
                minlength=4,
            ),
            html.Hr(),
            html.Div(
                children=create_connection_extra_inputs(
                    con.connection_type if con is not None else def_con_type, con
                ),
                id=EXTRA_PROPERTY_CONTAINER,
            ),
            html.Hr(),
            html.Div(
                [
                    dbc.Button(
                        [html.B(className="bi bi-check-circle"), " Test connection"],
                        id=CONNECTION_TEST_BUTTON,
                        color="primary",
                        class_name="me-3",
                    ),
                    create_delete_button(con),
                ],
                className="mb-3",
            ),
            dbc.Button(
                [dbc.Spinner(size="sm"), " Loading..."],
                color="light",
                id=TEST_CONNECTION_LOADING,
                disabled=True,
                size="sm",
                class_name="border-1",
                style={"display": "none"},
            ),
            html.Div(children=[], id=TEST_CONNECTION_RESULT),
            html.Div(children=[], id=DELETE_CONNECTION_RESULT),
            create_confirm_dialog(),
        ],
    )


@callback(
    Output(EXTRA_PROPERTY_CONTAINER, "children"),
    Input({"type": INDEX_TYPE, "index": PROP_CONNECTION_TYPE}, "value"),
    State(MITZU_LOCATION, "pathname"),
    prevent_initial_call=True,
)
@restricted
def connection_type_changed(connection_type: str, pathname: str) -> List[bc.Component]:
    deps: DEPS.Dependencies = cast(
        DEPS.Dependencies, flask.current_app.config.get(DEPS.CONFIG_KEY)
    )
    if pathname == P.CONNECTIONS_CREATE_PATH:
        connection: Optional[M.Connection] = None
    else:
        connection_id = P.get_path_value(
            P.CONNECTIONS_MANAGE_PATH, pathname, P.CONNECTION_ID_PATH_PART
        )
        connection = deps.storage.get_connection(connection_id)
    con_type = M.ConnectionType.parse(connection_type)
    return create_connection_extra_inputs(con_type, connection)


@callback(
    Output(CONFIRM_DIALOG_ACCEPT, "n_clicks"),
    Input(CONFIRM_DIALOG_ACCEPT, "n_clicks"),
    State(MITZU_LOCATION, "pathname"),
    prevent_initial_call=True,
)
@restricted
def delete_confirmed_clicked(n_clicks: int, pathname: str) -> int:
    if n_clicks is None:
        return no_update
    deps: DEPS.Dependencies = cast(
        DEPS.Dependencies, flask.current_app.config.get(DEPS.CONFIG_KEY)
    )
    connection_id = P.get_path_value(
        P.CONNECTIONS_MANAGE_PATH, pathname, P.CONNECTION_ID_PATH_PART
    )
    deps.storage.delete_connection(connection_id)
    return n_clicks


@callback(
    Output(CONFIRM_DIALOG_INDEX, "is_open"),
    Output(DELETE_CONNECTION_RESULT, "children"),
    Input(CONNECTION_DELETE_BUTTON, "n_clicks"),
    Input(CONFIRM_DIALOG_CLOSE, "n_clicks"),
    State(MITZU_LOCATION, "pathname"),
    prevent_initial_call=True,
)
@restricted
def delete_button_clicked(delete: int, close: int, pathname: str) -> Tuple[bool, str]:
    if delete is None:
        return no_update, no_update
    deps: DEPS.Dependencies = cast(
        DEPS.Dependencies, flask.current_app.config.get(DEPS.CONFIG_KEY)
    )
    if ctx.triggered_id == CONNECTION_DELETE_BUTTON:
        connection_id = P.get_path_value(
            P.CONNECTIONS_MANAGE_PATH, pathname, P.CONNECTION_ID_PATH_PART
        )
        project_ids = deps.storage.list_projects()
        for p_id in project_ids:
            prj = deps.storage.get_project(p_id)
            if prj.get_connection_id() == connection_id:
                return False, html.Div(
                    children=[
                        html.Span(
                            "You can't delete this connection because it is used by  "
                        ),
                        dcc.Link(
                            prj.project_name,
                            P.create_path(P.PROJECTS_MANAGE_PATH, project_id=p_id),
                        ),
                    ],
                    className="my-3 text-danger lead",
                )
        return True, ""

    return False, ""


def validate_input_values(values: Dict[str, Any]) -> Optional[str]:
    for k, v in values.items():
        if k not in NOT_REQUIRED_PROPERTIES and (
            v is None or len(v) < MIN_LENGTH or len(v) > MAX_LENGTH
        ):
            return k
    return None


@callback(
    Output(TEST_CONNECTION_RESULT, "children"),
    Input(CONNECTION_TEST_BUTTON, "n_clicks"),
    State({"type": INDEX_TYPE, "index": ALL}, "value"),
    State(MITZU_LOCATION, "pathname"),
    prevent_initial_call=True,
    background=True,
    running=[
        (
            Output(TEST_CONNECTION_LOADING, "style"),
            {"display": "inline-block"},
            {"display": "none"},
        ),
    ],
    interval=100,
)
@restricted
def test_connection_clicked(
    n_clicks: int, values: Dict[str, Any], pathname: str
) -> List[bc.Component]:
    if n_clicks is None:
        return no_update

    vals = {}
    for prop in ctx.args_grouping[1]:
        id_val = prop["id"]
        if id_val.get("type") == INDEX_TYPE:
            vals[id_val.get("index")] = prop["value"]

    invalid = validate_input_values(values=vals)
    if invalid is not None:
        return html.P(f"Invalid {value_to_label(invalid)}", className="lead my-3")

    try:
        connection = create_connection_from_values(vals)
        dummy_project = M.Project(connection, [], project_name="dp")
        dummy_project.get_adapter().test_connection()
        return html.P("Connected successfully!", className="lead my-3")
    except Exception as exc:
        traceback.print_exc()
        return html.P(f"Failed to connect: {exc}", className="my-3 text-danger")



from __future__ import annotations

import logging
import os
import pathlib
from urllib import parse
from abc import ABC
from copy import copy
from dataclasses import dataclass, field
from datetime import datetime, timedelta
from enum import Enum, auto
from typing import (
    Any,
    Callable,
    Dict,
    Generic,
    List,
    Optional,
    Tuple,
    TypeVar,
    Union,
    cast,
    Protocol,
)
import pandas as pd

from dateutil import parser
from dateutil.relativedelta import relativedelta

import mitzu.adapters.adapter_factory as factory
import mitzu.adapters.generic_adapter as GA
import mitzu.helper as helper
import mitzu.notebook.model_loader as ML
import mitzu.project_discovery as D
import mitzu.visualization.charts as CHRT
import mitzu.visualization.plot as PLT
import mitzu.visualization.titles as TI
import mitzu.project_serialization as PSE


ANY_EVENT_NAME = "any_event"


class MetricType(Enum):
    SEGMENTATION = auto()
    CONVERSION = auto()
    RETENTION = auto()
    JOURNEY = auto()


class TimeGroup(Enum):
    TOTAL = auto()
    SECOND = auto()
    MINUTE = auto()
    HOUR = auto()
    DAY = auto()
    WEEK = auto()
    MONTH = auto()
    QUARTER = auto()
    YEAR = auto()

    @classmethod
    def parse(cls, val: Union[str, TimeGroup]) -> TimeGroup:
        if type(val) == TimeGroup:
            return val
        elif type(val) == str:
            val = val.upper()
            if val.endswith("S"):
                val = val[:-1]
            return TimeGroup[val]
        else:
            raise ValueError(f"Invalid argument type for TimeGroup parse: {type(val)}")

    def __str__(self) -> str:
        return self.name.lower()

    @staticmethod
    def group_by_string(tg: TimeGroup) -> str:
        if tg == TimeGroup.TOTAL:
            return "Overall"
        if tg == TimeGroup.SECOND:
            return "Every Second"
        if tg == TimeGroup.MINUTE:
            return "Every Minute"
        if tg == TimeGroup.HOUR:
            return "Hourly"
        if tg == TimeGroup.DAY:
            return "Daily"
        if tg == TimeGroup.WEEK:
            return "Weekly"
        if tg == TimeGroup.MONTH:
            return "Monthly"
        if tg == TimeGroup.QUARTER:
            return "Quarterly"
        if tg == TimeGroup.YEAR:
            return "Yearly"
        raise Exception("Unkonwn timegroup value exception")


class AggType(Enum):

    COUNT_UNIQUE_USERS = auto()
    COUNT_EVENTS = auto()
    CONVERSION = auto()
    RETENTION_RATE = auto()
    PERCENTILE_TIME_TO_CONV = auto()
    AVERAGE_TIME_TO_CONV = auto()

    def to_agg_str(self, agg_param: Any = None) -> str:
        if self == AggType.CONVERSION:
            return "conversion"
        if self == AggType.RETENTION_RATE:
            return "retention_rate"
        if self == AggType.COUNT_EVENTS:
            return "event_count"
        if self == AggType.COUNT_UNIQUE_USERS:
            return "user_count"
        if self == AggType.AVERAGE_TIME_TO_CONV:
            return "ttc_avg"
        if self == AggType.PERCENTILE_TIME_TO_CONV:
            if agg_param is None:
                raise ValueError(
                    "For percentile time to convert aggregation agg_param is required"
                )
            return f"ttc_p{agg_param:.0f}"
        raise ValueError(f"{self} is not supported for to str")

    @staticmethod
    def parse_agg_str(val: str) -> Tuple[AggType, Any]:
        val = val.lower()
        if val == "event_count":
            return (AggType.COUNT_EVENTS, None)
        if val == "user_count":
            return (AggType.COUNT_UNIQUE_USERS, None)
        if val.startswith("ttc_p") and val[5:].isnumeric():
            param = int(val[5:])
            if 0 < param > 100:
                raise ValueError("Percentile value must be an integer between 0 and 99")
            return (AggType.PERCENTILE_TIME_TO_CONV, param)
        if val == "ttc_median":
            return (AggType.PERCENTILE_TIME_TO_CONV, 50)
        if val == "ttc_avg":
            return (AggType.AVERAGE_TIME_TO_CONV, None)
        if val == "conversion":
            return (AggType.CONVERSION, None)
        if val == "retention_rate":
            return (AggType.RETENTION_RATE, None)
        raise ValueError(
            f"Unsupported AggType: {val}\n"
            "supported['event_count', 'user_count', 'ttc_median', 'ttc_p90', 'ttc_p95',"
            " 'ttc_avg', 'conversion', 'retention']"
        )


class SimpleChartType(Enum):
    BAR = auto()
    STACKED_BAR = auto()
    LINE = auto()
    STACKED_AREA = auto()
    HEATMAP = auto()

    @classmethod
    def parse(cls, value: Union[str, SimpleChartType]) -> SimpleChartType:
        if type(value) == SimpleChartType:
            return value
        elif type(value) == str:
            for key, enm in SimpleChartType._member_map_.items():
                if key.lower() == value.lower():
                    return SimpleChartType[key]
            raise ValueError(
                f"Unknown chart type {value} supported are {[k.lower() for k in SimpleChartType._member_names_]}"
            )
        else:
            raise ValueError(
                f"Parse should only be str or SimpleChartType but it was {type(value)}"
            )


class Operator(Enum):
    EQ = auto()
    NEQ = auto()
    GT = auto()
    LT = auto()
    GT_EQ = auto()
    LT_EQ = auto()
    ANY_OF = auto()
    NONE_OF = auto()
    LIKE = auto()
    NOT_LIKE = auto()
    IS_NULL = auto()
    IS_NOT_NULL = auto()

    def __str__(self) -> str:
        if self == Operator.EQ:
            return "="
        if self == Operator.NEQ:
            return "!="
        if self == Operator.GT:
            return ">"
        if self == Operator.LT:
            return "<"
        if self == Operator.GT_EQ:
            return ">="
        if self == Operator.LT_EQ:
            return "<="
        if self == Operator.ANY_OF:
            return "any of"
        if self == Operator.NONE_OF:
            return "none of"
        if self == Operator.LIKE:
            return "like"
        if self == Operator.NOT_LIKE:
            return "not like"
        if self == Operator.IS_NULL:
            return "is null"
        if self == Operator.IS_NOT_NULL:
            return "is not null"

        raise ValueError(f"Not supported operator for title: {self}")


class BinaryOperator(Enum):
    AND = auto()
    OR = auto()

    def __str__(self) -> str:
        return self.name.lower()


class DataType(Enum):
    STRING = auto()
    NUMBER = auto()
    BOOL = auto()
    DATETIME = auto()
    MAP = auto()
    STRUCT = auto()

    def is_complex(self) -> bool:
        return self in (DataType.MAP, DataType.STRUCT)

    def from_string(self, string_value: str) -> Any:
        if self == DataType.BOOL:
            return bool(string_value)
        if self == DataType.NUMBER:
            return float(string_value)
        if self == DataType.STRING:
            return string_value
        if self == DataType.DATETIME:
            try:
                return parser.parse(string_value)
            except parser.ParserError:
                return string_value
        else:
            raise Exception(f"Unsupported parsing for type: {self.name}.")


class AttributionMode(Enum):
    FIRST_EVENT = 1
    LAST_EVENT = 2
    ALL_EVENTS = 3


class ConnectionType(Enum):
    FILE = "file"
    ATHENA = "athena"
    TRINO = "trino"
    POSTGRESQL = "postgresql+psycopg2"
    MYSQL = "mysql+mysqlconnector"
    SQLITE = "sqlite"
    DATABRICKS = "databricks"
    SNOWFLAKE = "snowflake"

    @classmethod
    def parse(cls, val: str | ConnectionType) -> ConnectionType:
        if type(val) == str:
            val = val.upper()
            return ConnectionType[val]
        elif type(val) == ConnectionType:
            return val
        else:
            raise ValueError(
                f"Invalid argument type for ConnectionType parse: {type(val)}"
            )


@dataclass(frozen=True)
class TimeWindow:
    value: int = 1
    period: TimeGroup = TimeGroup.DAY

    @classmethod
    def parse(cls, val: str | TimeWindow) -> TimeWindow:
        if type(val) == str:
            vals = val.strip().split(" ")
            return TimeWindow(value=int(vals[0]), period=TimeGroup.parse(vals[1]))
        elif type(val) == TimeWindow:
            return val
        else:
            raise ValueError(f"Invalid argument type for TimeWindow parse: {type(val)}")

    def __str__(self) -> str:
        prular = "s" if self.value > 1 else ""
        return f"{self.value} {self.period}{prular}"

    def to_relative_delta(self) -> relativedelta:
        if self.period == TimeGroup.SECOND:
            return relativedelta(seconds=self.value)
        if self.period == TimeGroup.MINUTE:
            return relativedelta(minutes=self.value)
        if self.period == TimeGroup.HOUR:
            return relativedelta(hours=self.value)
        if self.period == TimeGroup.DAY:
            return relativedelta(days=self.value)
        if self.period == TimeGroup.WEEK:
            return relativedelta(weeks=self.value)
        if self.period == TimeGroup.MONTH:
            return relativedelta(months=self.value)
        if self.period == TimeGroup.QUARTER:
            return relativedelta(months=self.value * 4)
        if self.period == TimeGroup.YEAR:
            return relativedelta(year=self.value)
        raise Exception(f"Unsupported relative delta value: {self.period}")


class Identifiable(Protocol):
    def get_id(self) -> str:
        pass


T = TypeVar("T")
ID = TypeVar("ID", bound=Identifiable)


class InvalidReferenceException(Exception):
    pass


@dataclass
class State(Generic[T]):
    _val: Optional[T] = None

    def get_value(self) -> Optional[T]:
        return self._val

    def get_value_if_exsits(self) -> T:
        if self._val is None:
            raise ValueError("State is empty.")
        return self._val

    def set_value(self, value: Optional[T]):
        self._val = value

    def __getstate__(self):
        """Override so pickle doesn't store state"""
        return {}

    def __setstate__(self, state):
        """Override so pickle doesn't store state"""


@dataclass(init=False)
class Reference(Generic[ID]):
    """We need the Reference class to support multitenant Objects. E.g.
    If the Connection or Project objects are reused we don't want to store them in the DB
    multiple times. The reference protected _id must be the id of the object
    """

    _value_state: State[ID]
    _id: Optional[str]

    def __init__(self, value: Optional[ID]):
        self._value_state = State(value)
        if value is None:
            self._id = None
        else:
            self._id = value.get_id()

    def get_id(self) -> Optional[str]:
        return self._id

    def get_value(self) -> Optional[ID]:
        return self._value_state.get_value()

    def get_value_if_exists(self) -> ID:
        res = self.get_value()
        if res is None:
            raise InvalidReferenceException(
                f"Missing reference value for id: {self._id}"
            )
        return res

    def set_value(self, value: Optional[ID]):
        self._value_state.set_value(value)
        if value is None:
            self._id = None
        else:
            self._id = value.get_id()

    def restore_value(self, value: Optional[ID]):
        if self._id is None:
            raise InvalidReferenceException("Restoring reference without ID.")
        self._value_state.set_value(value)
        if value is not None and value.get_id() != self._id:
            raise InvalidReferenceException(
                "Restored reference value has different ID."
            )


class SecretResolver(ABC):
    """
    Abstract base class for all secret resolvers
    """

    def resolve_secret(self) -> str:
        raise NotImplementedError()


@dataclass(frozen=True)
class PromptSecretResolver(SecretResolver):
    """
    Prompts for a secret value on the first query execution in an interactive Python environment.

    :param title: the title of the prompt
    """

    title: str = "Secret"

    def resolve_secret(self) -> str:
        import getpass

        return getpass.getpass(prompt=self.title)


@dataclass(frozen=True)
class ConstSecretResolver(SecretResolver):
    """
    Resolves a secret with a preconfigured static value.

    :param secret: the secret value
    """

    secret: str

    def resolve_secret(self) -> str:
        return self.secret


@dataclass(frozen=True)
class EnvVarSecretResolver(SecretResolver):
    """
    Resolves a secret from an environmental variable or raises an exception if the environment variable is not found.

    :param variable_name: name of the environment variable with containing the secret
    """

    variable_name: str

    def resolve_secret(self) -> str:
        secret = os.getenv(self.variable_name)
        if secret is not None:
            return secret
        else:
            raise Exception(f"Environmental variable {self.variable_name} was not set.")


@dataclass(frozen=True)
class Connection(Identifiable):
    """
    Contains all details needed to connect to a data warehouse.

    :param connection_type: type of the connection
    :param user_name: username
    :param secret_resolver: secret resolver to get the user password
    :param host: hostname
    :param port: port number
    :param schema: schema name
    :param catalog: catalog name
    :param url_params: extra sqlite connection url parameters
    :param extra_configs: used for connection adapter configuration
    """

    connection_name: str
    connection_type: ConnectionType
    id: str = field(default_factory=helper.create_unique_id)
    user_name: Optional[str] = None
    host: Optional[str] = None
    port: Optional[int] = None
    url: Optional[str] = None
    # TBD remove schema
    schema: Optional[str] = None
    catalog: Optional[str] = None
    # Used for connection url parametrization
    url_params: Optional[str] = None
    # Used for adapter configuration
    extra_configs: Dict[str, Any] = field(default_factory=dict)
    _secret: State[str] = field(default_factory=State)
    secret_resolver: Optional[SecretResolver] = None

    @property
    def password(self):
        if self._secret.get_value() is None:
            if self.secret_resolver is not None:
                secret = self.secret_resolver.resolve_secret()
                self._secret.set_value(secret)
            else:
                return ""
        return self._secret.get_value()

    def get_url_param(self, param: str) -> Optional[str]:
        res = parse.parse_qs(self.url_params).get(param)
        if res is not None and len(res) == 1:
            return res[0]
        return None

    def get_id(self) -> str:
        return self.id


@dataclass(frozen=True)
class DiscoverySettings:
    max_enum_cardinality: int = 300
    max_map_key_cardinality: int = 1000

    end_dt: Optional[datetime] = None
    property_sample_rate: int = 0

    lookback_days: int = 14
    min_property_sample_size: int = 1000


class WebappEndDateConfig(Enum):
    CUSTOM_DATE = auto()
    NOW = auto()
    START_OF_CURRENT_DAY = auto()
    END_OF_CURRENT_DAY = auto()


@dataclass(frozen=True)
class WebappSettings:
    lookback_window: TimeWindow = TimeWindow(30, TimeGroup.DAY)
    auto_refresh_enabled: bool = True
    end_date_config: WebappEndDateConfig = WebappEndDateConfig.START_OF_CURRENT_DAY
    custom_end_date: Optional[datetime] = None


class InvalidEventDataTableError(Exception):
    pass


@dataclass(unsafe_hash=False)
class EventDataTable(Identifiable):
    """
    Refers to a single table in the data warehouse or data lake.
    """

    table_name: str
    event_time_field: Field
    user_id_field: Field

    id: str = field(default_factory=helper.create_unique_id)
    schema: Optional[str] = None
    catalog: Optional[str] = None
    event_name_field: Optional[Field] = None
    date_partition_field: Optional[Field] = None
    event_name_alias: Optional[str] = None

    # TBD change to Field type
    ignored_fields: List[Field] = field(default_factory=lambda: [])

    # TBD change to Field type
    event_specific_fields: Optional[List[Field]] = None

    # TBD remove
    description: Optional[str] = None

    discovery_settings: Optional[DiscoverySettings] = None
    project_reference: Reference[Project] = field(
        default_factory=lambda: Reference(None)
    )

    @classmethod
    def create(
        cls,
        table_name: str,
        event_time_field: Union[str, Field],
        user_id_field: Union[str, Field],
        schema: Optional[str] = None,
        catalog: Optional[str] = None,
        event_name_field: Optional[Union[str, Field]] = None,
        event_name_alias: Optional[str] = None,
        ignored_fields: Optional[List[Union[str, Field]]] = None,
        event_specific_fields: List[Union[str, Field]] = None,
        date_partition_field: Optional[Union[str, Field]] = None,
        description: str = None,
        discovery_settings: Optional[DiscoverySettings] = None,
    ):

        if event_name_field == "":
            event_name_field = None
        if not event_name_field and event_name_alias is None:
            event_name_alias = table_name

        if type(event_name_field) == str:
            event_name_field = Field(_name=event_name_field, _type=DataType.STRING)

        if type(date_partition_field) == str:
            date_partition_field = Field(
                _name=date_partition_field, _type=DataType.DATETIME
            )

        if type(event_time_field) == str:
            event_time_field = Field(_name=event_time_field, _type=DataType.DATETIME)

        if type(user_id_field) == str:
            user_id_field = Field(_name=user_id_field, _type=DataType.STRING)

        converted_ignored_fields: List[Field] = []
        if ignored_fields is not None:
            for igf in ignored_fields:
                if type(igf) == str:
                    converted_ignored_fields.append(
                        Field(_name=igf, _type=DataType.STRING)
                    )
                elif isinstance(igf, Field):
                    converted_ignored_fields.append(igf)

        converted_evt_spc_fields: List[Field] = []
        if event_specific_fields is not None:
            for esf in event_specific_fields:
                if type(esf) == str:
                    if "." in esf:
                        raise ValueError(f"Invalid field name {esf}")
                    converted_evt_spc_fields.append(
                        Field(_name=esf, _type=DataType.STRING)
                    )
                elif isinstance(esf, Field):
                    converted_evt_spc_fields.append(esf)

        return EventDataTable(
            table_name=table_name,
            event_name_alias=event_name_alias,
            description=description,
            ignored_fields=converted_ignored_fields,
            event_specific_fields=converted_evt_spc_fields,
            event_name_field=cast(Field, event_name_field),
            date_partition_field=cast(Field, date_partition_field),
            event_time_field=cast(Field, event_time_field),
            user_id_field=cast(Field, user_id_field),
            schema=schema,
            catalog=catalog,
            discovery_settings=discovery_settings,
        )

    @classmethod
    def single_event_table(
        cls,
        table_name: str,
        event_time_field: str,
        user_id_field: str,
        schema: Optional[str] = None,
        catalog: Optional[str] = None,
        event_name_alias: Optional[str] = None,
        ignored_fields: List[Union[str, Field]] = None,
        date_partition_field: Optional[str] = None,
        description: str = None,
        discovery_settings: Optional[DiscoverySettings] = None,
    ):
        """
        Creates an Event Data Table from a table in a data warehouse

        :param table_name: name of the table
        :param event_time_field: name of the field containing the event time
        :param user_id_field: name of the field containing the user ID
        :param schema:
        :param catalog:
        :param event_name_alias:
        :param ignored_fields: name of the field which should be ignored
        :param date_partition_field: name of the field used for partitioning the data by date
        :param description:
        :param discovery_settings: discovery settings, if None then the project wide discovery settings will be used
        """
        return EventDataTable.create(
            table_name=table_name,
            event_name_alias=event_name_alias,
            description=description,
            ignored_fields=ignored_fields,
            event_specific_fields=None,
            event_name_field=None,
            date_partition_field=date_partition_field,
            event_time_field=event_time_field,
            user_id_field=user_id_field,
            schema=schema,
            catalog=catalog,
            discovery_settings=discovery_settings,
        )

    @classmethod
    def multi_event_table(
        cls,
        table_name: str,
        event_time_field: str,
        user_id_field: str,
        schema: Optional[str] = None,
        catalog: Optional[str] = None,
        event_name_field: Optional[str] = None,
        ignored_fields: List[Union[str, Field]] = None,
        event_specific_fields: List[Union[str, Field]] = None,
        date_partition_field: Optional[str] = None,
        description: str = None,
        discovery_settings: Optional[DiscoverySettings] = None,
    ):
        """
        Creates an Event Data Table from a table in a data warehouse

        :param table_name: name of the table
        :param event_time_field: name of the field containing the event time
        :param user_id_field: name of the field containing the user ID
        :param schema:
        :param catalog:
        :param event_name_alias:
        :param ignored_fields: name of the field which should be ignored
        :param event_specific_fields: name of the fields which are specific for certain events,
            these fields will be discovered separately for every event.
        :param date_partition_field: name of the field used for partitioning the data by date
        :param description:
        :param discovery_settings: discovery settings, if None then the project wide discovery settings will be used
        """
        return EventDataTable.create(
            table_name=table_name,
            event_name_alias=None,
            description=description,
            ignored_fields=ignored_fields,
            event_specific_fields=event_specific_fields,
            event_name_field=event_name_field,
            date_partition_field=date_partition_field,
            event_time_field=event_time_field,
            user_id_field=user_id_field,
            schema=schema,
            catalog=catalog,
            discovery_settings=discovery_settings,
        )

    def __hash__(self):
        return self.id.__hash__()

    def __eq__(self, other):
        if isinstance(other, EventDataTable):
            return self.id == other.id
        return False

    def get_id(self) -> str:
        return self.id

    def get_full_name(self) -> str:
        schema = "" if self.schema is None else self.schema + "."
        catalog = "" if self.catalog is None else self.catalog + "."
        return f"{catalog}{schema}{self.table_name}"

    def update_discovery_settings(self, discovery_settings: DiscoverySettings):
        properties = self.__dict__.copy()
        properties.update(discovery_settings=discovery_settings)
        return EventDataTable(**properties)

    def validate(self, adapter: GA.GenericDatasetAdapter):
        if self.event_name_alias is not None and self.event_name_field is not None:
            raise InvalidEventDataTableError(
                f"For {self.table_name} both event_name_alias and event_name_field can't be defined in the same time."
            )
        if self.event_name_alias is None and self.event_name_field is None:
            raise InvalidEventDataTableError(
                f"For {self.table_name} define the event_name_alias or the event_name_field property."
            )

        available_fields = [f._get_name().lower() for f in adapter.list_fields(self)]
        if len(available_fields) == 0:
            raise InvalidEventDataTableError(
                f"No fields in event data table '{self.table_name}'"
            )

        for field_to_validate in [
            self.event_name_field,
            self.event_time_field,
            self.user_id_field,
            self.date_partition_field,
        ]:
            if field_to_validate is None:
                continue

            if field_to_validate._get_name().lower() not in available_fields:
                raise InvalidEventDataTableError(
                    f"Event data table '{self.table_name}' does not have '{field_to_validate._get_name()}' field"
                )

    @property
    def project(self) -> Project:
        res = self.project_reference.get_value()
        if res is None:
            raise InvalidReferenceException("EventDataTable doesn't have a Project")
        return res

    def set_project(self, project: Project):
        self.project_reference = Reference(project)


class InvalidProjectError(Exception):
    pass


@dataclass(init=False)
class Project(Identifiable):
    """
    Defines a Mitzu project

    :param connection: configurations to connect to a data warehouse
    :param event_data_tables: list of Event Data Tables containing the events
    :param discovery_settings: default, project wide discovery settings
    :param webapp_settings: configurations to represent the project in the Mitzu webapp
    """

    project_name: str
    event_data_tables: List[EventDataTable]
    discovery_settings: DiscoverySettings
    webapp_settings: WebappSettings
    description: Optional[str]
    id: str
    _connection_ref: Reference[Connection]
    _adapter_cache: State[GA.GenericDatasetAdapter]
    _discovered_project: State[DiscoveredProject]

    def __init__(
        self,
        connection: Connection,
        event_data_tables: List[EventDataTable],
        project_name: str,
        project_id: Optional[str] = None,
        description: Optional[str] = None,
        discovery_settings: Optional[DiscoverySettings] = None,
        webapp_settings: Optional[WebappSettings] = None,
    ):
        object.__setattr__(self, "id", project_id)
        edt_with_discovery_settings = []
        if discovery_settings is None:
            discovery_settings = DiscoverySettings()
        for edt in event_data_tables:
            edt.set_project(self)
            if edt.discovery_settings is None:
                edt_with_discovery_settings.append(
                    edt.update_discovery_settings(discovery_settings)
                )
            else:
                edt_with_discovery_settings.append(edt)

        if webapp_settings is None:
            webapp_settings = WebappSettings()

        if project_id is None:
            project_id = helper.create_unique_id()

        object.__setattr__(self, "event_data_tables", edt_with_discovery_settings)
        object.__setattr__(self, "discovery_settings", discovery_settings)
        object.__setattr__(self, "webapp_settings", webapp_settings)
        object.__setattr__(self, "project_name", project_name)
        object.__setattr__(self, "description", description)
        object.__setattr__(self, "_connection_ref", Reference(connection))
        object.__setattr__(self, "_adapter_cache", State())
        object.__setattr__(self, "_discovered_project", State())

    def get_adapter(self) -> GA.GenericDatasetAdapter:
        val = self._adapter_cache.get_value()
        if val is None:
            adp = factory.create_adapter(self)
            self._adapter_cache.set_value(adp)
            return adp
        else:
            return val

    @property
    def connection(self) -> Connection:
        res = self._connection_ref.get_value()
        if res is None:
            raise InvalidReferenceException("Connection is missing from the Project")
        return res

    def get_connection_id(self) -> str:
        res = self._connection_ref.get_id()
        if res is None:
            raise InvalidReferenceException("Project has not Connection ID")
        return res

    def set_connection(self, connection: Optional[Connection]):
        self._connection_ref.set_value(connection)

    def restore_connection(self, connection: Optional[Connection]):
        self._connection_ref.restore_value(connection)

    def get_default_end_dt(self) -> datetime:
        if (
            self.webapp_settings.end_date_config == WebappEndDateConfig.CUSTOM_DATE
            and self.webapp_settings.custom_end_date is not None
        ):
            return self.webapp_settings.custom_end_date
        if (
            self.webapp_settings.end_date_config
            == WebappEndDateConfig.START_OF_CURRENT_DAY
        ):
            return datetime.now().replace(hour=0, minute=0, second=0, microsecond=0)
        if (
            self.webapp_settings.end_date_config
            == WebappEndDateConfig.END_OF_CURRENT_DAY
        ):
            return datetime.now().replace(
                hour=0, minute=0, second=0, microsecond=0
            ) + timedelta(days=1)
        return datetime.now()

    def get_default_discovery_start_dt(self) -> datetime:
        return self.get_default_end_dt() - timedelta(
            days=self.discovery_settings.lookback_days
        )

    def discover_project(
        self, progress_bar: bool = True, callback: Optional[Callable] = None
    ) -> DiscoveredProject:
        """
        Discovers all Event Data Tables with the given discovery settings.

        :param progress_bar: if True then a progressbar will be shown
        :param callback: callback function for each EDT discovered
        :return: DiscoveredProject containing all the discovered event properties
        """
        return D.ProjectDiscovery(project=self, callback=callback).discover_project(
            progress_bar
        )

    def validate(self):
        if len(self.event_data_tables) == 0:
            raise InvalidProjectError(
                "At least a single EventDataTable needs to be added to the Project.\n"
                "Project(event_data_tables = [ EventDataTable.create(...)])"
            )
        try:
            self.get_adapter().test_connection()
        except Exception as e:
            raise InvalidProjectError(f"Connection failed: {str(e)}") from e

        for edt in self.event_data_tables:
            edt.validate(self.get_adapter())

    def get_id(self) -> str:
        return self.id


class DatasetModel:
    @classmethod
    def _to_globals(cls, glbs: Dict):
        for k, v in cls.__dict__.items():
            if k != "_to_globals":
                glbs[k] = v


@dataclass(frozen=True, init=False)
class DiscoveredProject:
    definitions: Dict[EventDataTable, Dict[str, Reference[EventDef]]]
    project: Project

    def __init__(
        self,
        definitions: Dict[EventDataTable, Dict[str, Reference[EventDef]]],
        project: Project,
    ) -> None:
        object.__setattr__(self, "definitions", definitions)
        object.__setattr__(self, "project", project)
        project._discovered_project.set_value(self)

    def __post_init__(self):
        self.project.validate()

    def create_notebook_class_model(self) -> Any:
        return ML.ModelLoader().create_datasource_class_model(self)

    def display_inline_dashboard(
        self,
        port: int = 8080,
        host: str = "0.0.0.0",
        external_url: Optional[str] = None,
        logging_level: int = logging.WARN,
        height: Any = 800,
        width: Any = "100%",
    ):
        import IPython.display as ID

        if not external_url:
            external_url = f"http://{host}:{port}"
        results = self.display_external_dashboard(
            port, host, logging_level, new_thread=True
        )
        ID.display(ID.IFrame(external_url, height=height, width=width))
        return results

    def display_external_dashboard(
        self,
        port: Optional[int] = 8080,
        host: Optional[str] = "0.0.0.0",
        logging_level: int = logging.WARN,
        new_thread: bool = False,
    ):
        import mitzu.webapp.notebook as DASH

        self.project.validate()
        results: Optional[Dict[str, Any]] = {}
        DASH.external_dashboard(
            self,
            results=results,
            port=port,
            host=host,
            logging_level=logging_level,
            new_thread=new_thread,
        )
        return results

    def get_event_def(self, event_name) -> EventDef:
        for val in self.definitions.values():
            res = val.get(event_name)
            if res is not None:
                return res.get_value_if_exists()
        raise Exception(
            f"Invalid state, {event_name} is not present in Discovered Datasource."
        )

    def get_all_events(self) -> Dict[str, EventDef]:
        res: Dict[str, EventDef] = {}
        for val in self.definitions.values():
            res = {**res, **{k: v.get_value_if_exists() for k, v in val.items()}}
        return res

    def get_all_event_names(self) -> List[str]:
        res: List[str] = []
        for val in self.definitions.values():
            res.extend(val.keys())
        return res

    @staticmethod
    def _get_path(
        project_name: str, folder: str = "./", extension="mitzu"
    ) -> pathlib.Path:
        if project_name.endswith(f".{extension}"):
            return pathlib.Path(folder, f"{project_name}")
        else:
            return pathlib.Path(folder, f"{project_name}.{extension}")

    def save_to_project_file(
        self, project_name: str, folder: str = "./", extension="mitzu"
    ):
        path = self._get_path(project_name, folder, extension)
        with path.open(mode="w") as file:
            file.write(PSE.serialize_discovered_project(self))

    @classmethod
    def load_from_project_file(
        cls, project_name: str, folder: str = "./", extension="mitzu"
    ) -> DiscoveredProject:
        path = cls._get_path(project_name, folder, extension)
        with path.open(mode="rb") as file:
            return PSE.deserialize_discovered_project(file.read())


@dataclass(frozen=True, init=False)
class Field:
    _name: str
    _type: DataType = field(repr=False)
    _sub_fields: Optional[Tuple[Field, ...]] = field(hash=False, compare=False)
    _parent: Optional[Field] = field(
        repr=False,
        default=None,
    )

    def __init__(
        self,
        _name: str,
        _type: DataType,
        _sub_fields: Optional[Tuple[Field, ...]] = None,
    ):
        object.__setattr__(self, "_name", _name)
        object.__setattr__(self, "_type", _type)
        object.__setattr__(self, "_sub_fields", _sub_fields)
        if _sub_fields is not None:
            for sf in _sub_fields:
                object.__setattr__(sf, "_parent", self)

    def has_sub_field(self, field: Field) -> bool:
        if self._sub_fields is None:
            return False
        curr = field
        while curr._parent is not None:
            curr = curr._parent
        return curr == self

    def get_all_subfields(self) -> List[Field]:
        if self._sub_fields is None:
            return [self]
        else:
            all_sfs = []
            for sf in self._sub_fields:
                all_sfs.extend(sf.get_all_subfields())
            return all_sfs

    def __str__(self) -> str:
        if self._sub_fields is not None:
            return "(" + (", ".join([str(f) for f in self._sub_fields])) + ")"
        return f"{self._name} {self._type.name}"

    def _get_name(self) -> str:
        if self._parent is None:
            return self._name
        return f"{self._parent._get_name()}.{self._name}"


@dataclass(frozen=True)
class EventFieldDef:
    _event_name: str
    _field: Field
    _event_data_table: EventDataTable
    _description: Optional[str] = ""
    _enums: Optional[List[Any]] = None

    @property
    def _project(self) -> Project:
        return self._event_data_table.project


@dataclass(frozen=True)
class EventDef(Identifiable):
    _event_name: str
    _fields: Dict[Field, EventFieldDef]
    _event_data_table: EventDataTable
    _description: Optional[str] = ""

    @property
    def _project(self) -> Project:
        return self._event_data_table.project

    def get_id(self) -> str:
        return self._event_data_table.id + f"_{self._event_name}"


# =========================================== Metric definitions ===========================================

DEF_MAX_GROUP_COUNT = 10
DEF_LOOK_BACK_DAYS = TimeWindow(30, TimeGroup.DAY)
DEF_CONV_WINDOW = TimeWindow(1, TimeGroup.DAY)
DEF_RET_WINDOW = TimeWindow(1, TimeGroup.WEEK)
DEF_TIME_GROUP = TimeGroup.DAY


class Resolution(Enum):
    EVERY_EVENT = auto()
    ONE_USER_EVENT_PER_MINUTE = auto()
    ONE_USER_EVENT_PER_HOUR = auto()
    ONE_USER_EVENT_PER_DAY = auto()

    @classmethod
    def parse(cls, val: str | Resolution) -> Resolution:
        if type(val) == str:
            val = val.upper()
            return Resolution[val]
        elif type(val) == Resolution:
            return val
        else:
            raise ValueError(f"Invalid argument type for Resolution parse: {type(val)}")

    def get_time_group(self) -> Optional[TimeGroup]:
        if self == Resolution.ONE_USER_EVENT_PER_DAY:
            return TimeGroup.DAY
        if self == Resolution.ONE_USER_EVENT_PER_HOUR:
            return TimeGroup.HOUR
        if self == Resolution.ONE_USER_EVENT_PER_MINUTE:
            return TimeGroup.MINUTE
        return None


@dataclass()
class MetricConfig:
    start_dt: Optional[datetime] = None
    end_dt: Optional[datetime] = None
    lookback_days: TimeWindow = DEF_LOOK_BACK_DAYS
    time_group: TimeGroup = DEF_TIME_GROUP
    max_group_count: int = DEF_MAX_GROUP_COUNT
    group_by: Optional[EventFieldDef] = None
    custom_title: Optional[str] = None
    agg_type: Optional[AggType] = None
    agg_param: Optional[Any] = None
    chart_type: Optional[SimpleChartType] = None
    resolution: Resolution = Resolution.EVERY_EVENT


@dataclass(init=False, frozen=True)
class Metric(ABC):
    _config: MetricConfig

    def __init__(self, config: MetricConfig):
        object.__setattr__(self, "_config", config)

    @property
    def _max_group_count(self) -> int:
        if self._config.max_group_count is None:
            return DEF_MAX_GROUP_COUNT
        return self._config.max_group_count

    @property
    def _lookback_days(self) -> TimeWindow:
        return self._config.lookback_days

    @property
    def _time_group(self) -> TimeGroup:
        return self._config.time_group

    @property
    def _resolution(self) -> Resolution:
        return self._config.resolution

    @property
    def _group_by(self) -> Optional[EventFieldDef]:
        return self._config.group_by

    @property
    def _chart_type(self) -> Optional[SimpleChartType]:
        return self._config.chart_type

    @property
    def _custom_title(self) -> Optional[str]:
        return self._config.custom_title

    @property
    def _end_dt(self) -> datetime:
        if self._config.end_dt is not None:
            return self._config.end_dt
        return self.get_project().get_default_end_dt()

    @property
    def _start_dt(self) -> datetime:
        if self._config.start_dt is not None:
            return self._config.start_dt
        return self._end_dt - self._lookback_days.to_relative_delta()

    @property
    def _agg_type(self) -> AggType:
        if self._config.agg_type is not None:
            return self._config.agg_type
        if isinstance(self, ConversionMetric):
            return AggType.CONVERSION
        if isinstance(self, SegmentationMetric):
            return AggType.COUNT_UNIQUE_USERS
        if isinstance(self, RetentionMetric):
            return AggType.RETENTION_RATE
        raise NotImplementedError(
            f"_agg_type property is not implemented for {type(self)}"
        )

    @property
    def _agg_param(self) -> Any:
        if self._config.agg_param is not None:
            return self._config.agg_param
        if self._agg_type == AggType.PERCENTILE_TIME_TO_CONV:
            return 50
        return None

    def get_project(self) -> Project:
        raise NotImplementedError()

    def get_df(self) -> pd.DataFrame:
        raise NotImplementedError()

    def get_sql(self) -> pd.DataFrame:
        raise NotImplementedError()

    def get_title(self) -> str:
        raise NotImplementedError()

    def print_sql(self):
        print(self.get_sql())

    def get_figure(self):
        chart = CHRT.get_simple_chart(self)
        return PLT.plot_chart(chart, self)

    def __repr__(self) -> str:
        fig = self.get_figure()
        fig.show(config={"displayModeBar": False})
        return ""


class ConversionMetric(Metric):
    _conversion: Conversion
    _conv_window: TimeWindow

    def __init__(
        self,
        conversion: Conversion,
        config: MetricConfig,
        conv_window: TimeWindow = DEF_CONV_WINDOW,
    ):
        super().__init__(config)
        self._conversion = conversion
        self._conv_window = conv_window

    def get_df(self) -> pd.DataFrame:
        project = helper.get_segment_project(self._conversion._segments[0])
        return project.get_adapter().get_conversion_df(self)

    def get_sql(self) -> pd.DataFrame:
        project = helper.get_segment_project(self._conversion._segments[0])
        return project.get_adapter().get_conversion_sql(self)

    def get_figure(self):
        chart = CHRT.get_simple_chart(self)
        return PLT.plot_chart(chart, self)

    def __repr__(self) -> str:
        return super().__repr__()

    def get_project(self) -> Project:
        curr: Segment = self._conversion._segments[0]
        while not isinstance(curr, SimpleSegment):
            curr = cast(ComplexSegment, curr)._left
        return curr._left._project

    def get_title(self) -> str:
        return TI.get_conversion_title(self)


@dataclass(frozen=True, init=False)
class SegmentationMetric(Metric):
    _segment: Segment

    def __init__(self, segment: Segment, config: MetricConfig):
        super().__init__(config)
        object.__setattr__(self, "_segment", segment)

    def get_df(self) -> pd.DataFrame:
        project = helper.get_segment_project(self._segment)
        return project.get_adapter().get_segmentation_df(self)

    def get_sql(self) -> str:
        project = helper.get_segment_project(self._segment)
        return project.get_adapter().get_segmentation_sql(self)

    def __repr__(self) -> str:
        return super().__repr__()

    def get_project(self) -> Project:
        curr: Segment = self._segment
        while not isinstance(curr, SimpleSegment):
            curr = cast(ComplexSegment, curr)._left
        return curr._left._project

    def get_title(self) -> str:
        return TI.get_segmentation_title(self)


@dataclass(frozen=True, init=False)
class RetentionMetric(Metric):

    _initial_segment: Segment
    _retaining_segment: Segment
    _retention_window: TimeWindow

    def __init__(
        self,
        initial_segment: Segment,
        retaining_segment: Segment,
        retention_window: TimeWindow,
        config: MetricConfig,
    ):
        super().__init__(config)
        object.__setattr__(self, "_initial_segment", initial_segment)
        object.__setattr__(self, "_retaining_segment", retaining_segment)
        object.__setattr__(self, "_retention_window", retention_window)

    def config(
        self,
        start_dt: Optional[str | datetime] = None,
        end_dt: Optional[str | datetime] = None,
        custom_title: Optional[str] = None,
        retention_window: Union[TimeWindow, str] = TimeWindow(
            value=1, period=TimeGroup.WEEK
        ),
        time_group: Union[str, TimeGroup] = TimeGroup.WEEK,
        group_by: Optional[EventFieldDef] = None,
        max_group_by_count: int = DEF_MAX_GROUP_COUNT,
        lookback_days: Union[int, TimeWindow] = DEF_LOOK_BACK_DAYS,
        chart_type: Optional[Union[str, SimpleChartType]] = None,
        resolution: Union[str, Resolution] = Resolution.EVERY_EVENT,
    ) -> RetentionMetric:
        chart_type = chart_type
        if type(chart_type) == str:
            chart_type = SimpleChartType.parse(chart_type)

        config = MetricConfig(
            start_dt=helper.parse_datetime_input(start_dt, None),
            end_dt=helper.parse_datetime_input(end_dt, None),
            time_group=(
                time_group
                if type(time_group) == TimeGroup
                else TimeGroup.parse(time_group)
            ),
            custom_title=custom_title,
            group_by=group_by,
            max_group_count=max_group_by_count,
            lookback_days=(
                lookback_days
                if type(lookback_days) == TimeWindow
                else TimeWindow(cast(int, lookback_days), TimeGroup.DAY)
            ),
            agg_type=AggType.RETENTION_RATE,
            agg_param=None,
            resolution=(
                resolution
                if type(resolution) == Resolution
                else Resolution.parse(resolution)
            ),
            chart_type=cast(SimpleChartType, chart_type),
        )

        return RetentionMetric(
            initial_segment=self._initial_segment,
            retaining_segment=self._retaining_segment,
            retention_window=(
                retention_window
                if type(retention_window) == TimeWindow
                else TimeWindow.parse(retention_window)
            ),
            config=config,
        )

    def __repr__(self) -> str:
        return super().__repr__()

    def get_df(self) -> pd.DataFrame:
        project = helper.get_segment_project(self._initial_segment)
        return project.get_adapter().get_retention_df(self)

    def get_sql(self) -> str:
        project = helper.get_segment_project(self._initial_segment)
        return project.get_adapter().get_retention_sql(self)

    def get_project(self) -> Project:
        curr: Segment = self._initial_segment
        while not isinstance(curr, SimpleSegment):
            curr = cast(ComplexSegment, curr)._left
        return curr._left._project

    def get_title(self) -> str:
        return TI.get_retention_title(self)


class Conversion(ConversionMetric):
    def __init__(self, segments: List[Segment]):
        super().__init__(self, config=MetricConfig())
        self._segments = segments

    def __rshift__(self, right: Segment) -> Conversion:
        segments = copy(self._segments)
        segments.append(right)
        return Conversion(segments)

    def config(
        self,
        conv_window: Optional[Union[str, TimeWindow]] = DEF_CONV_WINDOW,
        start_dt: Optional[Union[str, datetime]] = None,
        end_dt: Optional[Union[str, datetime]] = None,
        time_group: Union[str, TimeGroup] = DEF_TIME_GROUP,
        group_by: Optional[EventFieldDef] = None,
        max_group_by_count: int = DEF_MAX_GROUP_COUNT,
        lookback_days: Union[int, TimeWindow] = DEF_LOOK_BACK_DAYS,
        custom_title: Optional[str] = None,
        aggregation: Union[str, AggType] = AggType.CONVERSION,
        chart_type: Optional[Union[str, SimpleChartType]] = None,
        resolution: Union[str, Resolution] = Resolution.EVERY_EVENT,
    ) -> ConversionMetric:
        if type(lookback_days) == int:
            lookback_days = TimeWindow(lookback_days, TimeGroup.DAY)

        agg_param = None
        if type(aggregation) != AggType:
            agg_type, agg_param = AggType.parse_agg_str(cast(str, aggregation))
        else:
            agg_type = aggregation

        chart_type = chart_type
        if type(chart_type) == str:
            chart_type = SimpleChartType.parse(chart_type)

        config = MetricConfig(
            start_dt=helper.parse_datetime_input(start_dt, None),
            end_dt=helper.parse_datetime_input(end_dt, None),
            time_group=(
                time_group
                if type(time_group) == TimeGroup
                else TimeGroup.parse(time_group)
            ),
            custom_title=custom_title,
            group_by=group_by,
            max_group_count=max_group_by_count,
            lookback_days=(
                lookback_days
                if type(lookback_days) == TimeWindow
                else TimeWindow(cast(int, lookback_days), TimeGroup.DAY)
            ),
            agg_type=agg_type,
            agg_param=agg_param,
            resolution=(
                resolution
                if type(resolution) == Resolution
                else Resolution.parse(resolution)
            ),
            chart_type=cast(SimpleChartType, chart_type),
        )

        if conv_window is not None:
            conv_res = ConversionMetric(conversion=self._conversion, config=config)
            conv_res._conv_window = TimeWindow.parse(conv_window)
            return conv_res
        else:
            raise ValueError("conw_window or ret_window must be defined")

    def __repr__(self) -> str:
        return super().__repr__()


class Segment(SegmentationMetric):
    def __init__(self):
        super().__init__(self, config=MetricConfig())

    def __and__(self, right: Segment) -> ComplexSegment:
        return ComplexSegment(self, BinaryOperator.AND, right)

    def __or__(self, right: Segment) -> ComplexSegment:
        return ComplexSegment(self, BinaryOperator.OR, right)

    def __rshift__(self, right: Segment) -> Conversion:
        return Conversion([self, right])

    def __ge__(self, retaining_segment: Segment) -> RetentionMetric:
        return RetentionMetric(
            self,
            retaining_segment,
            retention_window=TimeWindow(value=1, period=TimeGroup.WEEK),
            config=MetricConfig(time_group=TimeGroup.WEEK),
        )

    def config(
        self,
        start_dt: Optional[Union[str, datetime]] = None,
        end_dt: Optional[Union[str, datetime]] = None,
        time_group: Union[str, TimeGroup] = DEF_TIME_GROUP,
        group_by: Optional[EventFieldDef] = None,
        max_group_by_count: int = DEF_MAX_GROUP_COUNT,
        lookback_days: Union[int, TimeWindow] = DEF_LOOK_BACK_DAYS,
        custom_title: Optional[str] = None,
        aggregation: Union[str, AggType] = AggType.COUNT_UNIQUE_USERS,
        chart_type: Optional[Union[str, SimpleChartType]] = None,
    ) -> SegmentationMetric:
        agg_param = None
        if type(aggregation) != AggType:
            agg_type, agg_param = AggType.parse_agg_str(cast(str, aggregation))
        else:
            agg_type = aggregation

        chart_type = chart_type
        if type(chart_type) == str:
            chart_type = SimpleChartType.parse(chart_type)

        config = MetricConfig(
            start_dt=helper.parse_datetime_input(start_dt, None),
            end_dt=helper.parse_datetime_input(end_dt, None),
            time_group=(
                time_group
                if type(time_group) == TimeGroup
                else TimeGroup.parse(time_group)
            ),
            custom_title=custom_title,
            group_by=group_by,
            max_group_count=max_group_by_count,
            lookback_days=(
                lookback_days
                if type(lookback_days) == TimeWindow
                else TimeWindow(cast(int, lookback_days), TimeGroup.DAY)
            ),
            agg_type=agg_type,
            agg_param=agg_param,
            resolution=Resolution.EVERY_EVENT,
            chart_type=cast(SimpleChartType, chart_type),
        )

        return SegmentationMetric(segment=self, config=config)

    def __repr__(self) -> str:
        return super().__repr__()


@dataclass(init=False, frozen=True)
class ComplexSegment(Segment):
    _left: Segment
    _operator: BinaryOperator
    _right: Segment

    def __init__(self, _left: Segment, _operator: BinaryOperator, _right: Segment):
        object.__setattr__(self, "_left", _left)
        object.__setattr__(self, "_operator", _operator)
        object.__setattr__(self, "_right", _right)
        super().__init__()

    def __repr__(self) -> str:
        return super().__repr__()

    def __hash__(self) -> int:
        return hash(f"{hash(self._left)}{self._operator}{hash(self._right)}")


@dataclass(init=False, frozen=True)
class SimpleSegment(Segment):
    _left: EventFieldDef | EventDef  # str is an event_name without any filters
    _operator: Optional[Operator] = None
    _right: Optional[Any] = None

    def __init__(
        self,
        _left: EventFieldDef | EventDef,
        _operator: Optional[Operator] = None,
        _right: Optional[Any] = None,
    ):
        object.__setattr__(self, "_left", _left)
        object.__setattr__(self, "_operator", _operator)
        object.__setattr__(self, "_right", _right)
        super().__init__()

    def __repr__(self) -> str:
        return super().__repr__()

    def __hash__(self) -> int:
        event_property_name = (
            self._left._event_name
            if type(self._left) != EventFieldDef
            else f"{self._left._event_name}.{self._left._field._get_name()}"
        )
        return hash(f"{event_property_name}{self._operator}{self._right}")



import mitzu.webapp.configs as configs
import redis
import diskcache
from typing import Any, Optional, List, Dict
from abc import ABC
from dataclasses import dataclass
import pickle
import mitzu.helper as H
import flask


class MitzuCache(ABC):
    def put(self, key: str, val: Any, expire: Optional[float] = None) -> None:
        """Puts some data to the storage

        Args:
            key (str): the key of the data
            val (Any): the picklable value
            expire (Optional[float], optional): seconds until expiration

        Raises:
            NotImplementedError: _description_
        """
        raise NotImplementedError()

    def get(self, key: str, default: Optional[Any] = None) -> Any:
        raise NotImplementedError()

    def clear(self, key: str) -> None:
        raise NotImplementedError()

    def clear_all(self, prefix: Optional[str] = None) -> None:
        for key in self.list_keys(prefix):
            self.clear(key)

    def list_keys(
        self, prefix: Optional[str] = None, strip_prefix: bool = True
    ) -> List[str]:
        raise NotImplementedError()


@dataclass(frozen=True, init=False)
class DiskMitzuCache(MitzuCache):

    _disk_cache: diskcache.Cache

    def __init__(self, name: str) -> None:
        super().__init__()
        object.__setattr__(
            self,
            "_disk_cache",
            diskcache.Cache(timeout=10, directory=f"{configs.DISK_CACHE_PATH}/{name}"),
        )

    def __post_init__(self):
        H.LOGGER.info("Vacuuming disk cache")
        self._disk_cache._sql("vacuum")

    def put(self, key: str, val: Any, expire: Optional[float] = None):
        self.clear(key)
        if val is not None:
            H.LOGGER.debug(f"PUT: {key}: {type(val)}")
            self._disk_cache.add(key, value=val, expire=expire)
        else:
            H.LOGGER.debug(f"PUT: {key}: None")

    def get(self, key: str, default: Optional[Any] = None) -> Any:
        res = self._disk_cache.get(key)
        if res is not None:
            H.LOGGER.debug(f"GET: {key}: {type(res)}")
            return res
        else:
            H.LOGGER.debug(f"GET: {key}: None")
            return default

    def clear(self, key: str) -> None:
        H.LOGGER.debug(f"Clear {key}")
        self._disk_cache.pop(key)

    def list_keys(
        self, prefix: Optional[str] = None, strip_prefix: bool = True
    ) -> List[str]:
        keys = self._disk_cache.iterkeys()
        start_pos = len(prefix) if strip_prefix and prefix is not None else 0
        res = [k[start_pos:] for k in keys if prefix is None or k.startswith(prefix)]
        if H.LOGGER.getEffectiveLevel() == H.logging.DEBUG:
            H.LOGGER.debug(f"LIST {prefix}: {res}")
        return res

    def get_disk_cache(self) -> diskcache.Cache:
        return self._disk_cache


@dataclass(frozen=True)
class RequestCache(MitzuCache):
    """This cache is in-memory however it is ephemeral, it only stores values until the end of the request
    This is because it is not picklable.
    """

    delegate: MitzuCache

    def _get_request_cache(self) -> Dict[str, Any]:
        if flask.has_app_context():
            if "request_cache" not in flask.g:
                flask.g.request_cache = {}
            return flask.g.get("request_cache")
        return {}

    def put(self, key: str, val: Any, expire: Optional[float] = None):
        self.delegate.put(key, val, expire)

    def get(self, key: str, default: Optional[Any] = None) -> Any:
        res = self._get_request_cache().get(key)
        if res is None:
            res = self.delegate.get(key, default)

        return res

    def clear(self, key: str) -> None:
        cache = self._get_request_cache()
        if key in cache:
            cache.pop(key)
        self.delegate.clear(key)

    def list_keys(
        self, prefix: Optional[str] = None, strip_prefix: bool = True
    ) -> List[str]:
        return self.delegate.list_keys(prefix, strip_prefix)


class RedisException(Exception):
    pass


@dataclass(init=False, frozen=True)
class RedisMitzuCache(MitzuCache):

    _redis: redis.Redis

    def __init__(self, redis_cache: Optional[redis.Redis] = None) -> None:
        super().__init__()

        if redis_cache is not None:
            object.__setattr__(self, "_redis", redis_cache)
        else:
            if configs.STORAGE_REDIS_HOST is None:
                raise ValueError(
                    "STORAGE_REDIS_HOST env variable is not set, can't create redis cache."
                )
            object.__setattr__(
                self,
                "_redis",
                redis.Redis(
                    host=configs.STORAGE_REDIS_HOST,
                    port=configs.STORAGE_REDIS_PORT,
                    password=configs.STORAGE_REDIS_PASSWORD,
                ),
            )

    def put(self, key: str, val: Any, expire: Optional[float] = None):
        pickled_value = pickle.dumps(val)
        if H.LOGGER.getEffectiveLevel() == H.logging.DEBUG:
            H.LOGGER.debug(f"PUT: {key}: {len(pickled_value)}")
        res = self._redis.set(name=key, value=pickled_value, ex=expire)
        if not res:
            raise RedisException(f"Couldn't set {key}")

    def get(self, key: str, default: Optional[Any] = None) -> Any:
        res = self._redis.get(name=key)
        if res is None:
            H.LOGGER.debug(f"GET: {key}: None")
            if default is not None:
                return default
            return None
        if H.LOGGER.getEffectiveLevel() == H.logging.DEBUG:
            H.LOGGER.debug(f"GET: {key}: {len(res)}")
        return pickle.loads(res)

    def clear(self, key: str) -> None:
        H.LOGGER.debug(f"CLEAR: {key}")
        self._redis.delete(key)

    def list_keys(
        self, prefix: Optional[str] = None, strip_prefix: bool = True
    ) -> List[str]:
        if not prefix:
            prefix = ""
        keys = self._redis.keys(f"{prefix}*")
        start_pos = len(prefix) if strip_prefix else 0
        res = [k.decode()[start_pos:] for k in keys]
        if H.LOGGER.getEffectiveLevel() == H.logging.DEBUG:
            H.LOGGER.debug(f"LIST prefix={prefix}: {res}")
        return res



# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from mitzu.adapters.sqlalchemy.athena.sqlalchemy.error import AthenaQueryError

from sqlalchemy.sql import compiler

# https://docs.aws.amazon.com/athena/latest/ug/reserved-words.html
ATHENA_RESERVED_WORDS = {
    "ALL",
    "ALTER",
    "AND",
    "ARRAY",
    "AS",
    "AUTHORIZATION",
    "BETWEEN",
    "BIGINT",
    "BINARY",
    "BOOLEAN",
    "BOTH",
    "BY",
    "CASE",
    "CASHE",
    "CAST",
    "CHAR",
    "COLUMN",
    "CONF",
    "CONSTRAINT",
    "COMMIT",
    "CREATE",
    "CROSS",
    "CUBE",
    "CURRENT",
    "CURRENT_DATE",
    "CURRENT_TIMESTAMP",
    "CURSOR",
    "DATABASE",
    "DATE",
    "DAYOFWEEK",
    "DECIMAL",
    "DELETE",
    "DESCRIBE",
    "DISTINCT",
    "DOUBLE",
    "DROP",
    "ELSE",
    "END",
    "EXCHANGE",
    "EXISTS",
    "EXTENDED",
    "EXTERNAL",
    "EXTRACT",
    "FALSE",
    "FETCH",
    "FLOAT",
    "FLOOR",
    "FOLLOWING",
    "FOR",
    "FOREIGN",
    "FROM",
    "FULL",
    "FUNCTION",
    "GRANT",
    "GROUP",
    "GROUPING",
    "HAVING",
    "IF",
    "IMPORT",
    "IN",
    "INNER",
    "INSERT",
    "INT",
    "INTEGER",
    "INTERSECT",
    "INTERVAL",
    "INTO",
    "IS",
    "JOIN",
    "LATERAL",
    "LEFT",
    "LESS",
    "LIKE",
    "LOCAL",
    "MACRO",
    "MAP",
    "MORE",
    "NONE",
    "NOT",
    "NULL",
    "NUMERIC",
    "OF",
    "ON",
    "ONLY",
    "OR",
    "ORDER",
    "OUT",
    "OUTER",
    "OVER",
    "PARTIALSCAN",
    "PARTITION",
    "PERCENT",
    "PRECEDING",
    "PRECISION",
    "PRESERVE",
    "PRIMARY",
    "PROCEDURE",
    "RANGE",
    "READS",
    "REDUCE",
    "REGEXP",
    "REFERENCES",
    "REVOKE",
    "RIGHT",
    "RLIKE",
    "ROLLBACK",
    "ROLLUP",
    "ROW",
    "ROWS",
    "SELECT",
    "SET",
    "SMALLINT",
    "START,TABLE",
    "TABLESAMPLE",
    "THEN",
    "TIME",
    "TIMESTAMP",
    "TO",
    "TRANSFORM",
    "TRIGGER",
    "TRUE",
    "TRUNCATE",
    "UNBOUNDED,UNION",
    "UNIQUEJOIN",
    "UPDATE",
    "USER",
    "USING",
    "UTC_TIMESTAMP",
    "VALUES",
    "VARCHAR",
    "VIEWS",
    "WHEN",
    "WHERE",
    "WINDOW",
    "WITH",
    "ALTER",
    "AND",
    "AS",
    "BETWEEN",
    "BY",
    "CASE",
    "CAST",
    "CONSTRAINT",
    "CREATE",
    "CROSS",
    "CUBE",
    "CURRENT_DATE",
    "CURRENT_PATH",
    "CURRENT_TIME",
    "CURRENT_TIMESTAMP",
    "CURRENT_USER",
    "DEALLOCATE",
    "DELETE",
    "DESCRIBE",
    "DISTINCT",
    "DROP",
    "ELSE",
    "END",
    "ESCAPE",
    "EXCEPT",
    "EXECUTE",
    "EXISTS",
    "EXTRACT",
    "FALSE",
    "FIRST",
    "FOR",
    "FROM",
    "FULL",
    "GROUP",
    "GROUPING",
    "HAVING",
    "IN",
    "INNER",
    "INSERT",
    "INTERSECT",
    "INTO",
    "IS",
    "JOIN",
    "LAST",
    "LEFT",
    "LIKE",
    "LOCALTIME",
    "LOCALTIMESTAMP",
    "NATURAL",
    "NORMALIZE",
    "NOT",
    "NULL",
    "OF",
    "ON",
    "OR",
    "ORDER",
    "OUTER",
    "PREPARE",
    "RECURSIVE",
    "RIGHT",
    "ROLLUP",
    "SELECT",
    "TABLE",
    "THEN",
    "TRUE",
    "UNESCAPE",
    "UNION",
    "UNNEST",
    "USING",
    "VALUES",
    "WHEN",
    "WHERE",
    "WITH",
}


class AthenaSQLCompiler(compiler.SQLCompiler):
    def limit_clause(self, select, **kw):
        if select._offset_clause is not None:
            raise AthenaQueryError("OFFSET is not supported")
        if select._limit_clause is not None:
            return "\nLIMIT " + self.process(select._limit_clause, **kw)
        return ""


class AthenaDDLCompiler(compiler.DDLCompiler):
    pass


class AthenaTypeCompiler(compiler.GenericTypeCompiler):
    def visit_DOUBLE(self, type_, **kw):
        return "DOUBLE"

    def visit_NUMERIC(self, type_, **kw):
        return self.visit_DECIMAL(type_, **kw)

    def visit_NCHAR(self, type_, **kw):
        return self.visit_VARCHAR(type_, **kw)

    def visit_NVARCHAR(self, type_, **kw):
        return self.visit_VARCHAR(type_, **kw)

    def visit_TEXT(self, type_, **kw):
        return self.visit_VARCHAR(type_, **kw)

    def visit_BINARY(self, type_, **kw):
        return self.visit_VARBINARY(type_, **kw)

    def visit_CLOB(self, type_, **kw):
        return self.visit_VARCHAR(type_, **kw)

    def visit_NCLOB(self, type_, **kw):
        return self.visit_VARCHAR(type_, **kw)

    def visit_BLOB(self, type_, **kw):
        return self.visit_VARCHAR(type_, **kw)

    def visit_DATETIME(self, type_, **kw):
        return self.visit_TIMESTAMP(type_, **kw)


class AthenaIdentifierPreparer(compiler.IdentifierPreparer):
    reserved_words = ATHENA_RESERVED_WORDS



from typing import Any, List, Optional, cast

import dash.development.base_component as bc
import dash_bootstrap_components as dbc
import flask
from dash import (
    ALL,
    Input,
    Output,
    State,
    callback,
    ctx,
    html,
    no_update,
    register_page,
)

import mitzu.helper as H
import mitzu.model as M
import mitzu.webapp.dependencies as DEPS
import mitzu.webapp.navbar as NB
import mitzu.webapp.pages.connections.manage_connections_component as MCC
import mitzu.webapp.pages.paths as P
from mitzu.webapp.auth.decorator import restricted, restricted_layout
from mitzu.webapp.helper import MITZU_LOCATION

CONNECTION_SAVE_BUTTON = "connection_save_button"
CONNECTION_CLOSE_BUTTON = "connection_close_button"
SAVE_RESPONSE_CONTAINER = "save_response_container"


@restricted_layout
def no_connection_layout():
    return layout(None)


@restricted_layout
def layout(connection_id: Optional[str] = None, **query_params) -> bc.Component:
    connection: Optional[M.Connection] = None
    if connection_id is not None:
        depenednecies: DEPS.Dependencies = cast(
            DEPS.Dependencies, flask.current_app.config.get(DEPS.CONFIG_KEY)
        )
        connection = depenednecies.storage.get_connection(connection_id)

    title = "Create new connection" if connection is None else "Manage connection"

    return dbc.Form(
        [
            NB.create_mitzu_navbar("create-connection-navbar", []),
            dbc.Container(
                children=[
                    html.H4(title),
                    html.Hr(),
                    MCC.create_manage_connection_component(connection),
                    html.Hr(),
                    html.Div(
                        [
                            dbc.Button(
                                [html.B(className="bi bi-x"), " Close"],
                                color="secondary",
                                class_name="me-3",
                                id=CONNECTION_CLOSE_BUTTON,
                                href=P.CONNECTIONS_PATH,
                            ),
                            dbc.Button(
                                [html.B(className="bi bi-check-circle"), " Save"],
                                color="success",
                                id=CONNECTION_SAVE_BUTTON,
                            ),
                        ],
                        className="mb-3",
                    ),
                    html.Div(children=[], id=SAVE_RESPONSE_CONTAINER),
                ],
                class_name="mb-3",
            ),
        ]
    )


register_page(
    __name__ + "_create",
    path=P.CONNECTIONS_CREATE_PATH,
    title="Mitzu - Create Connection",
    layout=no_connection_layout,
)


register_page(
    __name__,
    path_template=P.CONNECTIONS_MANAGE_PATH,
    title="Mitzu - Manage Connection",
    layout=layout,
)


@callback(
    Output(SAVE_RESPONSE_CONTAINER, "children"),
    Input(CONNECTION_SAVE_BUTTON, "n_clicks"),
    State({"type": MCC.INDEX_TYPE, "index": ALL}, "value"),
    State(MITZU_LOCATION, "pathname"),
    prevent_initial_call=True,
)
@restricted
def save_button_clicked(
    n_clicks: int, values: List[Any], pathname: str
) -> List[bc.Component]:
    if n_clicks is None:
        return no_update

    vals = {}
    for prop in ctx.args_grouping[1]:
        id_val = prop["id"]
        if id_val.get("type") == MCC.INDEX_TYPE:
            vals[id_val.get("index")] = prop["value"]

    connection = MCC.create_connection_from_values(vals)
    depenednecies: DEPS.Dependencies = cast(
        DEPS.Dependencies, flask.current_app.config.get(DEPS.CONFIG_KEY)
    )

    invalid = MCC.validate_input_values(values=vals)
    if invalid is not None:
        return html.P(f"Invalid {H.value_to_label(invalid)}", className="lead")
    depenednecies.storage.set_connection(connection.id, connection)

    return [html.P("Connection saved", className="lead")]



from __future__ import annotations

from dataclasses import dataclass
from typing import Optional

from flask import Flask

import mitzu.webapp.auth.authorizer as A
import mitzu.webapp.cache as C
import mitzu.webapp.configs as configs
import mitzu.webapp.storage as S
import mitzu.webapp.service.user_service as U
import mitzu.webapp.service.events_service as E

CONFIG_KEY = "dependencies"


@dataclass(frozen=True)
class Dependencies:

    authorizer: Optional[A.OAuthAuthorizer]
    storage: S.MitzuStorage
    queue: C.MitzuCache
    cache: C.MitzuCache
    events_service: E.EventsService
    user_service: Optional[U.UserService] = None

    @classmethod
    def from_configs(cls, server: Flask) -> Dependencies:
        delegate_cache: C.MitzuCache
        if configs.STORAGE_REDIS_HOST is not None:
            delegate_cache = C.RedisMitzuCache()
        else:
            delegate_cache = C.DiskMitzuCache("cache")
        cache = C.RequestCache(delegate_cache)

        authorizer = None
        oauth_config = None
        user_service = None
        if configs.AUTH_BACKEND == "cognito":
            from mitzu.webapp.auth.cognito import Cognito

            oauth_config = Cognito.get_config()
        elif configs.AUTH_BACKEND == "google":
            from mitzu.webapp.auth.google import GoogleOAuth

            oauth_config = GoogleOAuth.get_config()

        elif configs.AUTH_BACKEND == "local":
            user_service = U.UserService(
                cache, root_password=configs.AUTH_ROOT_PASSWORD
            )

        if oauth_config or user_service:
            auth_config = A.AuthConfig(
                oauth=oauth_config,
                token_validator=A.JWTTokenValidator.create_from_oauth_config(
                    oauth_config
                )
                if oauth_config is not None
                else None,
                allowed_email_domain=configs.AUTH_ALLOWED_EMAIL_DOMAIN,
                token_signing_key=configs.AUTH_JWT_SECRET,
                session_timeout=configs.AUTH_SESSION_TIMEOUT,
                user_service=user_service,
            )
            authorizer = A.OAuthAuthorizer.create(auth_config)
            authorizer.setup_authorizer(server)

        queue: C.MitzuCache
        if configs.QUEUE_REDIS_HOST is not None:
            queue = C.RedisMitzuCache()
        else:
            queue = C.DiskMitzuCache("queue")

        # Adding cache layer over storage
        storage = S.MitzuStorage(cache)
        if configs.SETUP_SAMPLE_PROJECT:
            S.setup_sample_project(storage)

        events_service = E.EventsService(storage)

        return Dependencies(
            authorizer=authorizer,
            cache=cache,
            storage=storage,
            queue=queue,
            user_service=user_service,
            events_service=events_service,
        )



import traceback
from typing import List

import dash.development.base_component as bc
import dash_bootstrap_components as dbc

from dash import html, register_page, callback, Input, Output, ctx, State, ALL, dcc

import mitzu.webapp.dependencies as DEPS
import mitzu.webapp.model as WM
import mitzu.webapp.navbar as NB
import mitzu.webapp.pages.paths as P
from mitzu.webapp.auth.decorator import restricted, restricted_layout
from typing import cast
import flask
from dash_iconify import DashIconify

register_page(__name__, path_template=P.DASHBOARDS_PATH, title="Mitzu - Dashboards")


CREATE_NEW_DASHBOARD_BTN = "create_new_dashboard"
DASHBOARD_DELETE = "dashboard_delete"

SEPARATOR = "###"

CONFIRM_DIALOG_MODEL = "dashboard_delete_confirm"
CONFIRM_DIALOG_CLOSE = "dashboard_delete_confirm_dialog_close"
CONFIRM_DIALOG_ACCEPT = "dashboard_delete_confirm_dialog_accept"
CONFIRM_MODEL_BODY = "dashboard_confirm_modal_body"
CONFIRM_DASHBOARD_ID = "dashboard_confirm_id"

DASHBOARD_CONTAINER = "dashboard_container"
DASHBOARDS_INFO = "dashboard_info"


@restricted_layout
def layout(**query_params) -> bc.Component:
    storage = cast(
        DEPS.Dependencies, flask.current_app.config.get(DEPS.CONFIG_KEY)
    ).storage
    d_ids = storage.list_dashboards()
    dashboards = [storage.get_dashboard(d_id) for d_id in d_ids]

    return html.Div(
        [
            NB.create_mitzu_navbar("dashboard-navbar", []),
            dbc.Container(
                children=[
                    dbc.Row(
                        [
                            dbc.Col(
                                html.H4("Select a Dashboard", className="card-title"),
                                width="auto",
                            ),
                            dbc.Col(
                                dbc.Button(
                                    [
                                        html.B(
                                            className="bi bi-plus-circle me-1",
                                        ),
                                        "Add Dashboard",
                                    ],
                                    color="primary",
                                    href=P.DASHBOARDS_CREATE_PATH,
                                    id=CREATE_NEW_DASHBOARD_BTN,
                                ),
                                width="auto",
                                class_name="ms-auto",
                            ),
                        ]
                    ),
                    html.Hr(),
                    html.Div(children="", id=DASHBOARDS_INFO),
                    dbc.Row(
                        children=list_dashboards(dashboards),
                        id=DASHBOARD_CONTAINER,
                    ),
                    create_confirm_dialog(),
                ]
            ),
        ]
    )


def create_confirm_dialog():
    return dbc.Modal(
        [
            dbc.ModalBody(
                "Do you really want to delete this saved metric?",
                class_name="lead",
                id=CONFIRM_MODEL_BODY,
            ),
            dbc.ModalFooter(
                [
                    dbc.Button(
                        "Close",
                        id=CONFIRM_DIALOG_CLOSE,
                        size="sm",
                        color="secondary",
                        class_name="me-1",
                    ),
                    dbc.Button(
                        "Delete",
                        id=CONFIRM_DIALOG_ACCEPT,
                        size="sm",
                        color="danger",
                    ),
                ]
            ),
            html.Div(children="", className="d-none", id=CONFIRM_DASHBOARD_ID),
        ],
        id=CONFIRM_DIALOG_MODEL,
        is_open=False,
    )


def list_dashboards(dashboards: List[WM.Dashboard]) -> List[bc.Component]:
    children = []

    for dashboard in dashboards:
        try:
            comp = create_dashboard_selector(dashboard)
            children.append(dbc.Col(comp, sm=12, lg=4, class_name="mb-3"))
        except Exception as exc:
            traceback.print_exception(exc)

    if len(children) == 0:
        return html.H4(
            "You don't have any dashboards yet",
            className="card-title text-center",
        )

    return children


def create_dashboard_selector(dashboard: WM.Dashboard) -> bc.Component:
    details: List[str] = []
    if dashboard.created_at is not None:
        details.append(f"Created at: {dashboard.created_at}")
    if dashboard.last_updated_at is not None:
        details.append(f"Edited at: {dashboard.last_updated_at}")

    href = P.create_path(P.DASHBOARDS_EDIT_PATH, dashboard_id=dashboard.id)

    return dbc.Card(
        [
            dbc.Row(
                [
                    dbc.Col(
                        dcc.Link(
                            DashIconify(
                                icon="ic:twotone-dashboard",
                                width=100,
                                height=100,
                                color="gray",
                            ),
                            style={"width": "1rem", "height": "1rem"},
                            href=href,
                        ),
                        className="col-md-4 ps-3",
                    ),
                    dbc.Col(
                        dbc.CardBody(
                            [
                                html.Div(
                                    [
                                        dcc.Link(
                                            html.H6(
                                                dashboard.name,
                                                className="card-title",
                                            ),
                                            href=href,
                                            className="text-dark",
                                        ),
                                        dbc.Button(
                                            html.I(className="bi bi-x-lg"),
                                            color="dark",
                                            id={
                                                "type": DASHBOARD_DELETE,
                                                "index": f"{dashboard.id}{SEPARATOR}{dashboard.name}",
                                            },
                                            size="sm",
                                            outline=True,
                                            className="border-0 align-top",
                                        ),
                                    ],
                                    className="d-flex justify-content-between",
                                ),
                                dcc.Link(
                                    children=[
                                        html.P(
                                            # TODO add description
                                            "",
                                            className="card-text text-nowrap    ",
                                        ),
                                        html.Small(
                                            f"Created on {dashboard.created_at.strftime('%c')}",
                                            className="card-text text-muted",
                                        ),
                                    ],
                                    href=href,
                                    className="text-dark",
                                ),
                            ]
                        ),
                        className="col-md-8",
                    ),
                ],
                className="g-0 d-flex align-items-center",
            )
        ],
    )


@callback(
    Output(CONFIRM_DIALOG_MODEL, "is_open"),
    Output(CONFIRM_DASHBOARD_ID, "children"),
    Output(CONFIRM_MODEL_BODY, "children"),
    Input({"type": DASHBOARD_DELETE, "index": ALL}, "n_clicks"),
    Input(CONFIRM_DIALOG_CLOSE, "n_clicks"),
    Input(CONFIRM_DIALOG_ACCEPT, "n_clicks"),
    prevent_initial_call=True,
)
@restricted
def delete_button_clicked(delete: int, close: int, accept: int):
    if (
        ctx.triggered_id is None
        or ctx.triggered_id
        in [
            CONFIRM_DIALOG_CLOSE,
            CONFIRM_DIALOG_ACCEPT,
        ]
        or all(d is None for d in delete)
    ):
        return (False, "", "")

    index = ctx.triggered_id.get("index", "")
    parts = index.split(SEPARATOR)
    if len(parts) == 2:
        dashboard_id, name = tuple(parts)
        message = f"Do you really want to delete {name}"
        return (True, dashboard_id, message)

    return (False, "", "")


@callback(
    Output(DASHBOARD_CONTAINER, "children"),
    Input(CONFIRM_DIALOG_ACCEPT, "n_clicks"),
    State(CONFIRM_DASHBOARD_ID, "children"),
    background=True,
    running=[
        (
            Output(DASHBOARDS_INFO, "children"),
            [
                dbc.Spinner(
                    spinner_style={"width": "1rem", "height": "1rem"},
                    spinner_class_name="me-1",
                ),
                "Loading saved metrics",
            ],
            "",
        )
    ],
    interval=200,
    prevent_initial_call=True,
)
@restricted
def confirm_button_clicked(accept_button: int, dashboard_id: str) -> List:
    storage = cast(
        DEPS.Dependencies, flask.current_app.config.get(DEPS.CONFIG_KEY)
    ).storage
    if ctx.triggered_id is not None:
        storage.clear_dashboard(dashboard_id)

    d_ids = storage.list_dashboards()
    dashboards = [storage.get_dashboard(d_id) for d_id in d_ids]
    return list_dashboards(dashboards)



from typing import Dict, Optional, cast, List

import dash.development.base_component as bc
import dash_bootstrap_components as dbc
import dash_draggable as dd
from dash import ALL, Input, Output, callback, ctx, html, State, no_update, dcc
import mitzu.webapp.model as WM
import mitzu.webapp.dependencies as DEPS
import flask
import dash_mantine_components as dmc
import mitzu.visualization.plot as PLT
import mitzu.visualization.charts as CHRT
from dash_iconify import DashIconify
import mitzu.serialization as SE
import mitzu.webapp.pages.paths as P
from mitzu.webapp.auth.decorator import restricted
from urllib.parse import quote
from mitzu.webapp.helper import MISSING_RESOURCE_CSS

RESPONSIVE_GRID_LAYOUT = "responsive_grid_layout"

DEF_COLUMN_COUNT = 4
DEF_HEIGHT = 8
ADD_DASHBOARD_METRIC_BUTTON = "add_dashboard_metric"
DASHBOARD_ITEM_PREFIX = "dashboard_metric"
SAVED_METRICS_OFF_CANVAS = "dashboard_saved_metrics_off_canvas"
SAVED_METRICS_CONTAINER = "dashboard_saved_metrics_container"
SAVED_METRICS_SEARCH = "dashboard_saved_metrics_search"
DELETE_SAVED_METRICS_TYPE = "dashboard_delete_saved_metric"
ADD_SAVED_METRICS_TYPE = "dashboard_add_saved_metric"
DASHBOARD_NAME_INPUT = "dashboard_name_input"
DASHBOARD_ID = "dashboard_id"
DASHBOARD_REFRESH_BUTTON = "dashboard_refresh"
DASHBOARD_METRIC_GRAPH_TYPE = "dashboard_metric_graph_type"

DASHBOARD_SPINNER = "dashboard_spinner"


def create_layout_item(id: str, x: int, y: int, w: int, h: int) -> Dict[str, Dict]:
    default = {"i": id, "x": x, "y": y, "w": w, "h": h}
    small = {"i": id, "x": 0, "y": y, "w": 1, "h": h}
    return {"lg": default, "md": default, "sm": default, "xs": small, "xxs": small}


def create_grid_width(width: int) -> Dict:
    return {"lg": width, "md": width, "sm": width, "xs": 1, "xxs": 1}


def create_saved_metric_card(saved_metric: WM.SavedMetric) -> bc.Component:
    return dbc.Card(
        [
            html.A(
                children=[
                    dbc.Row(
                        [
                            dbc.Col(
                                dbc.CardImg(
                                    src=saved_metric.small_base64,
                                    className="img-fluid rounded-start",
                                    style={"width": "120px", "height": "100px"},
                                ),
                                className="col-md-4 ps-3",
                            ),
                            dbc.Col(
                                dbc.CardBody(
                                    [
                                        html.Div(
                                            [
                                                html.H5(
                                                    saved_metric.name,
                                                    className="card-title",
                                                ),
                                            ],
                                            className="d-flex justify-content-between",
                                        ),
                                        html.P(
                                            saved_metric.description,
                                            className="card-text",
                                        ),
                                        html.Small(
                                            saved_metric.created_at.strftime("%c"),
                                            className="card-text text-muted",
                                        ),
                                    ]
                                ),
                                className="col-md-8",
                            ),
                        ],
                        className="g-0 d-flex align-items-center",
                    )
                ],
                href="#",
                className="text-dark",
                id={"type": ADD_SAVED_METRICS_TYPE, "index": saved_metric.id},
            )
        ],
        class_name="mb-3 hover-shadow",
    )


def create_saved_metrics_off_canvas() -> bc.Component:
    return dbc.Offcanvas(
        children=[
            html.Hr(),
            dbc.Button(
                [html.B(className="bi bi-plus-circle me-1"), "Explore projects"],
                href=P.PROJECTS_PATH,
                target="_blank",
                size="small",
                color="light",
                class_name="w-100 mb-3",
            ),
            dmc.TextInput(
                id=SAVED_METRICS_SEARCH,
                placeholder="Search",
                debounce=200,
                className="mb-3",
                icon=DashIconify(icon="ic:baseline-search"),
            ),
            html.Div(
                children=[html.P("Loading saved metrics", className="lead")],
                id=SAVED_METRICS_CONTAINER,
            ),
        ],
        id=SAVED_METRICS_OFF_CANVAS,
        is_open=False,
        close_button=True,
        scrollable=True,
        placement="end",
        title="Add saved metric",
    )


def create_dashboard_metric_card(dm: WM.DashboardMetric) -> bc.Component:
    index = f"{DASHBOARD_ITEM_PREFIX}_{dm.get_saved_metric_id()}"
    if (
        dm.saved_metric is None
        or dm.saved_metric.metric is None
        or dm.saved_metric.project is None
    ):
        component = html.Div(
            "Missing chart",
            className="lead text-center pt-5 w-100 h-100 d-inline-block",
        )
        name = dm.saved_metric.name if dm.saved_metric is not None else "Missing chart"
        href = None
        edit_href = None
        missing_project = True
        border_class = "border border-2 border-warning"
    else:
        component = dcc.Graph(
            id={"type": DASHBOARD_METRIC_GRAPH_TYPE, "index": dm.get_saved_metric_id()},
            className="w-100 h-100",
            figure=PLT.plot_chart(dm.saved_metric.chart, dm.saved_metric.metric),
            config={"displayModeBar": False},
            responsive=True,
        )

        metric = dm.saved_metric.metric
        name = dm.saved_metric.name
        missing_project = dm.saved_metric.project is None
        if not missing_project:
            path = P.create_path(
                P.PROJECTS_EXPLORE_PATH, project_id=dm.saved_metric.get_project_id()
            )

            href = f"{path}?m={quote(SE.to_compressed_string(metric))}"
            edit_href = f"{path}?sm={dm.saved_metric.id}"
            border_class = ""
        else:
            border_class = MISSING_RESOURCE_CSS
            href = ""
            edit_href = ""

    return html.Div(
        html.Div(
            children=[
                component,
                html.Div(
                    [
                        html.Span(
                            name,
                            className="me-2 lead user-select-none",
                        ),
                        dmc.Menu(
                            children=[
                                dmc.MenuTarget(
                                    dmc.ActionIcon(
                                        DashIconify(
                                            icon="material-symbols:menu-rounded"
                                        ),
                                        size="lg",
                                        className="me-1",
                                    ),
                                ),
                                dmc.MenuDropdown(
                                    [
                                        dmc.MenuItem(
                                            "Explore",
                                            icon=DashIconify(
                                                icon="radix-icons:external-link"
                                            ),
                                            href=href,
                                            target="_blank",
                                            disabled=missing_project,
                                        ),
                                        dmc.MenuItem(
                                            "Edit",
                                            icon=DashIconify(icon="mdi:gear-outline"),
                                            href=edit_href,
                                            target="_blank",
                                            disabled=missing_project,
                                        ),
                                        dmc.MenuDivider(),
                                        dmc.MenuItem(
                                            "Delete",
                                            icon=DashIconify(icon="tabler:trash"),
                                            color="red",
                                            id={
                                                "type": DELETE_SAVED_METRICS_TYPE,
                                                "index": dm.get_saved_metric_id(),
                                            },
                                        ),
                                    ]
                                ),
                            ],
                            position="bottom-end",
                        ),
                    ],
                    className="position-absolute end-0 d-flex",
                    style={"top": "5px"},
                ),
            ],
            className="w-100 h-100",
        ),
        id=index,
        className="w-100 h-100 " + border_class,
    )


def create_responsive_layouts(dashboard: WM.Dashboard) -> Dict:
    layouts: Dict[str, List] = {"lg": [], "md": [], "sm": [], "xs": [], "xxs": []}
    for dm in dashboard.dashboard_metrics:
        lt = create_layout_item(
            id=f"{DASHBOARD_ITEM_PREFIX}_{dm.get_saved_metric_id()}",
            x=dm.x,
            y=dm.y,
            w=dm.width,
            h=dm.height,
        )
        for key, ref in lt.items():
            layouts[key].append(ref)
    return layouts


def create_manage_dashboard_container(
    dashboard: Optional[WM.Dashboard],
) -> bc.Component:
    if dashboard is None:
        dashboard = WM.Dashboard("New dashboard")

    layouts = create_responsive_layouts(dashboard)
    return html.Div(
        [
            create_saved_metrics_off_canvas(),
            dbc.Container(
                [
                    html.Div(
                        children=dashboard.id,
                        id=DASHBOARD_ID,
                        className="d-none",
                    ),
                    dbc.Row(
                        [
                            dbc.Col(
                                [
                                    dmc.TextInput(
                                        type="text",
                                        id=DASHBOARD_NAME_INPUT,
                                        placeholder="New dashboard",
                                        value=dashboard.name,
                                        required=True,
                                        className="name-input mb-3 d-inline-block",
                                        debounce=500,
                                    ),
                                ],
                                lg=4,
                                sm=12,
                            ),
                            dbc.Col(
                                width="auto",
                                class_name="ms-auto",
                                children=[
                                    dbc.Button(
                                        [
                                            html.B(
                                                className="bi bi-arrow-clockwise me-1"
                                            ),
                                            "Refresh",
                                        ],
                                        color="light",
                                        class_name="d-inline-block",
                                        id=DASHBOARD_REFRESH_BUTTON,
                                    ),
                                ],
                            ),
                            dbc.Col(
                                dbc.Button(
                                    [
                                        html.B(className="bi bi-plus-circle me-1"),
                                        "Add",
                                    ],
                                    color="primary",
                                    id=ADD_DASHBOARD_METRIC_BUTTON,
                                ),
                                width="auto",
                            ),
                        ]
                    ),
                    dbc.Progress(
                        id=DASHBOARD_SPINNER,
                        style={"height": "2px"},
                        color="primary",
                        value=0,
                        class_name="my-2",
                    ),
                    dd.ResponsiveGridLayout(
                        id=RESPONSIVE_GRID_LAYOUT,
                        nrows=1,
                        height=30,
                        save=False,
                        clearSavedLayout=True,
                        gridCols=create_grid_width(4),
                        layouts=layouts,
                        children=[
                            create_dashboard_metric_card(dm)
                            for dm in dashboard.dashboard_metrics
                        ],
                    ),
                ],
                class_name="mb-10",
            ),
        ],
    )


@callback(
    Output(SAVED_METRICS_OFF_CANVAS, "is_open"),
    Output(SAVED_METRICS_CONTAINER, "children"),
    Input(ADD_DASHBOARD_METRIC_BUTTON, "n_clicks"),
    Input(SAVED_METRICS_SEARCH, "value"),
    prevent_initial_call=True,
)
@restricted
def manage_saved_metrics_off_canvas(button: int, search_value: str):
    storage = cast(
        DEPS.Dependencies, flask.current_app.config.get(DEPS.CONFIG_KEY)
    ).storage
    sm_ids = storage.list_saved_metrics()
    saved_metrics = [storage.get_saved_metric(sm_id) for sm_id in sm_ids]
    if search_value:
        saved_metrics = [
            sm
            for sm in saved_metrics
            if (not search_value or (search_value.lower() in sm.name.lower()))
        ]

    if len(saved_metrics) > 0:
        children = [create_saved_metric_card(sm) for sm in saved_metrics]
    else:
        children = html.P("No results", className="lead")

    return True, children


@callback(
    Output(RESPONSIVE_GRID_LAYOUT, "children"),
    Output(RESPONSIVE_GRID_LAYOUT, "layouts"),
    Output(RESPONSIVE_GRID_LAYOUT, "clearSavedLayout"),
    Input({"type": ADD_SAVED_METRICS_TYPE, "index": ALL}, "n_clicks"),
    Input({"type": DELETE_SAVED_METRICS_TYPE, "index": ALL}, "n_clicks"),
    State(DASHBOARD_NAME_INPUT, "value"),
    State(DASHBOARD_ID, "children"),
    prevent_initial_call=True,
)
@restricted
def manage_dashboard_content(
    link_clicks: List[int],
    delete_clicks: List[int],
    dashboard_name: int,
    dashboard_id: str,
):
    if ctx.triggered_id is None:
        return no_update, no_update, no_update

    storage = cast(
        DEPS.Dependencies, flask.current_app.config.get(DEPS.CONFIG_KEY)
    ).storage

    if ctx.triggered_id.get("type") == DELETE_SAVED_METRICS_TYPE:
        delete_metric_id = ctx.triggered_id.get("index")
        if (
            delete_clicks is None
            or all([c is None for c in delete_clicks])
            or delete_clicks is None
        ):
            return no_update, no_update, no_update
        dashboard = storage.get_dashboard(dashboard_id)
        dashboard = dashboard.update(
            [
                dm
                for dm in dashboard.dashboard_metrics
                if dm.get_saved_metric_id() != delete_metric_id
            ]
        )

        storage.set_dashboard(dashboard.id, dashboard)
        children = [
            create_dashboard_metric_card(dm) for dm in dashboard.dashboard_metrics
        ]
        layouts = create_responsive_layouts(dashboard)
        return children, layouts, True
    if ctx.triggered_id.get("type") == ADD_SAVED_METRICS_TYPE:
        new_metric_id = ctx.triggered_id.get("index")
        if (
            link_clicks is None
            or all([c is None for c in link_clicks])
            or new_metric_id is None
        ):
            return no_update, no_update, no_update

        saved_metric = storage.get_saved_metric(new_metric_id)
        dashboard = storage.get_dashboard(dashboard_id)
        max_y = 0
        if dashboard is None:
            dashboard = WM.Dashboard(
                name=dashboard_name, id=dashboard_id, dashboard_metrics=[]
            )
        elif len(dashboard.dashboard_metrics) > 0:
            max_y = max(dashboard.dashboard_metrics, key=lambda dm: dm.y).y + DEF_HEIGHT

        new_dm = WM.DashboardMetric(
            saved_metric=saved_metric,
            x=0,
            y=max_y,
            width=2,
            height=DEF_HEIGHT,
        )
        for dm in dashboard.dashboard_metrics:
            if dm.get_saved_metric_id() == new_metric_id:
                return no_update, no_update, no_update

        dashboard.dashboard_metrics.append(new_dm)
        storage.set_dashboard(dashboard.id, dashboard)

        children = [
            create_dashboard_metric_card(dm) for dm in dashboard.dashboard_metrics
        ]
        layouts = create_responsive_layouts(dashboard)
        return children, layouts, True


@callback(
    Output(DASHBOARD_NAME_INPUT, "value"),
    Input(DASHBOARD_NAME_INPUT, "value"),
    State(DASHBOARD_ID, "children"),
    prevent_initial_call=True,
)
@restricted
def dashboard_name_changed(name: str, dashboard_id: str) -> str:
    if name is None:
        return no_update
    storage = cast(
        DEPS.Dependencies, flask.current_app.config.get(DEPS.CONFIG_KEY)
    ).storage

    dashboard = storage.get_dashboard(dashboard_id)
    if dashboard is not None:
        d_name = name if name else "Unnamed dashboard"
        dashboard = dashboard.rename(d_name)
        dashboard = storage.set_dashboard(dashboard_id, dashboard)
    return name


@callback(
    Output(DASHBOARD_ID, "children"),
    Input(RESPONSIVE_GRID_LAYOUT, "layouts"),
    State(DASHBOARD_ID, "children"),
    prevent_initial_call=True,
)
@restricted
def manage_layout_change(layouts: Dict[str, Dict], dashboard_id: str) -> str:
    if dashboard_id is None:
        return no_update

    storage = cast(
        DEPS.Dependencies, flask.current_app.config.get(DEPS.CONFIG_KEY)
    ).storage

    dashboard = storage.get_dashboard(dashboard_id)
    if dashboard is None:
        return no_update

    for lt in layouts["lg"]:
        prefix = len(DASHBOARD_ITEM_PREFIX) + 1
        id = lt["i"][prefix:]
        for dm in dashboard.dashboard_metrics:
            if dm.get_saved_metric_id() == id:
                dm.x = lt["x"]
                dm.y = lt["y"]
                dm.width = lt["w"]
                dm.height = lt["h"]

    dashboard = storage.set_dashboard(dashboard_id, dashboard)
    return no_update


@callback(
    Output({"type": DASHBOARD_METRIC_GRAPH_TYPE, "index": ALL}, "figure"),
    Output(DASHBOARD_SPINNER, "value"),
    Input(DASHBOARD_REFRESH_BUTTON, "n_clicks"),
    State(DASHBOARD_ID, "children"),
    prevent_initial_call=True,
    running=[
        (Output(RESPONSIVE_GRID_LAYOUT, "style"), {"opacity": "0.3"}, {"opacity": "1"}),
    ],
    progress=[
        Output(DASHBOARD_SPINNER, "value"),
    ],
    background=True,
)
@restricted
def refresh_dashboards(set_progress, refresh_button_click: int, dashboard_id: str):
    if dashboard_id is None or refresh_button_click is None:
        return no_update, 0

    storage = cast(
        DEPS.Dependencies, flask.current_app.config.get(DEPS.CONFIG_KEY)
    ).storage

    dashboard = storage.get_dashboard(dashboard_id)
    figures = []
    total = len(dashboard.dashboard_metrics)
    for index, dm in enumerate(dashboard.dashboard_metrics):
        if dm.saved_metric is not None:
            metric = dm.saved_metric.metric
            project = dm.saved_metric.project
            if metric is None or project is None:
                continue
            simple_chart = CHRT.get_simple_chart(metric)
            fig = PLT.plot_chart(simple_chart, metric)
            sm = WM.SavedMetric(
                metric=metric,
                name=dm.saved_metric.name,
                chart=simple_chart,
                project=project,
                image_base64=dm.saved_metric.small_base64,
                small_base64=dm.saved_metric.small_base64,
                id=dm.saved_metric.id,
            )
            storage.set_saved_metric(sm.id, sm)
            dm.set_saved_metric(sm)
            figures.append(fig)
        progress = int((index + 1) * 100.0 / total)
        set_progress(progress)

    storage.set_dashboard(dashboard_id, dashboard)
    return figures, 0



from typing import List, Optional, Tuple
from dataclasses import dataclass, field
from enum import Enum
import random
import string
from mitzu.helper import create_unique_id
import hashlib
import mitzu.webapp.cache as C
import mitzu.webapp.configs as configs


class Role(Enum):
    ADMIN = "admin"
    MEMBER = "member"

    @classmethod
    def all_values(cls):
        return [
            Role.ADMIN,
            Role.MEMBER,
        ]


@dataclass
class User:
    """
    Container class for describing a user
    """

    email: str
    password_hash: str
    password_salt: str
    id: str = field(default_factory=create_unique_id)
    role: Role = Role.MEMBER


class UserNotFoundException(Exception):
    """
    Raised when the user is not found in the local user store
    """

    def __init__(self):
        super().__init__("User not found")


class UserAlreadyExists(Exception):
    """
    Raised when the user already exists in the local user store
    """

    def __init__(self):
        super().__init__("User already exists")


class UserPasswordAndConfirmationDoNotMatch(Exception):
    """
    Raised when the password and the password confirmation do not match
    when adding a new user or when chaning the password
    """

    def __init__(self):
        super().__init__("Password and password confirmation do not match")


class RootUserCannotBeChanged(Exception):
    """
    Raised before the root user changes
    """

    def __init__(self):
        super().__init__("Root user cannot be changed")


class UserService:
    """
    UserService provides the a single API to manage users in the local user storage
    """

    def __init__(self, cache: C.MitzuCache, root_password: Optional[str] = None):
        self._cache = cache

        has_admin = False
        for user in self.list_users():
            if user.role == Role.ADMIN:
                has_admin = True
                break

        if root_password and not has_admin:
            self.new_user(
                configs.AUTH_ROOT_USER_EMAIL,
                root_password,
                root_password,
                role=Role.ADMIN,
            )

    def is_root_user(self, user_id: str) -> bool:
        user = self.get_user_by_id(user_id)
        return user is not None and user.email == configs.AUTH_ROOT_USER_EMAIL

    def list_users(self) -> List[User]:
        result = []
        for key in self._cache.list_keys("users."):
            result.append(self._cache.get(f"users.{key}"))
        return result

    def get_user_by_id(self, user_id: str) -> Optional[User]:
        return self._cache.get(f"users.{user_id}")

    def get_user_by_email(self, email: str) -> Optional[User]:
        for user in self.list_users():
            if user.email == email:
                return user
        return None

    def _store_user(self, user: User):
        key = f"users.{user.id}"
        if self._cache.get(key):
            self._cache.clear(key)
        self._cache.put(key, user)

    def update_password(self, user_id: str, password: str, password_confirmation: str):
        user = self.get_user_by_id(user_id)
        if user is None:
            raise UserNotFoundException()

        if password != password_confirmation:
            raise UserPasswordAndConfirmationDoNotMatch()

        hash, salt = self._get_password_hash_with_salt(password)
        user.password_hash = hash
        user.password_salt = salt
        self._store_user(user)

    def update_role(self, user_id: str, role: Role):
        user = self.get_user_by_id(user_id)
        if user is None:
            raise UserNotFoundException()

        if user.email == configs.AUTH_ROOT_USER_EMAIL:
            raise RootUserCannotBeChanged()

        user.role = role
        self._store_user(user)

    def _get_password_hash_with_salt(self, password: str) -> Tuple[str, str]:
        salt = "".join(random.choice(string.printable) for i in range(10))
        password_to_hash = f"{password}:{salt}"
        hash = hashlib.sha256(password_to_hash.encode()).hexdigest()
        return (hash, salt)

    def new_user(
        self,
        email: str,
        password: str,
        password_confirmation: str,
        role: Role = Role.MEMBER,
    ) -> str:
        if self.get_user_by_email(email) is not None:
            raise UserAlreadyExists()

        if password != password_confirmation:
            raise UserPasswordAndConfirmationDoNotMatch()

        hash, salt = self._get_password_hash_with_salt(password)

        user = User(
            email=email,
            password_hash=hash,
            password_salt=salt,
            role=role,
        )
        self._store_user(user)
        return user.id

    def get_user_by_email_and_password(
        self, email: str, password: str
    ) -> Optional[User]:
        user = self.get_user_by_email(email)
        if user is None:
            return None

        password_to_hash = f"{password}:{user.password_salt}"
        hash = hashlib.sha256(password_to_hash.encode()).hexdigest()
        if hash == user.password_hash:
            return user
        return None

    def delete_user(self, user_id: str):
        user = self.get_user_by_id(user_id)
        if user is None:
            raise UserNotFoundException()

        if user.email == configs.AUTH_ROOT_USER_EMAIL:
            raise RootUserCannotBeChanged()

        self._cache.clear(f"users.{user.id}")



PROJECT_ID_PATH_PART = "project_id"
CONNECTION_ID_PATH_PART = "connection_id"
USER_PATH_PART = "user_id"
DASHBOARD_ID = "dashboard_id"

ADMIN_PAGE = "/admin"

EVENTS_AND_PROPERTIES_PATH = "/events"
EVENTS_AND_PROPERTIES_PROJECT_PATH = f"/events/<{PROJECT_ID_PATH_PART}>"

PROJECTS_PATH = "/projects"
PROJECTS_CREATE_PATH = "/projects/create/"
PROJECTS_EXPLORE_PATH = f"/projects/<{PROJECT_ID_PATH_PART}>/explore"
PROJECTS_MANAGE_PATH = f"/projects/<{PROJECT_ID_PATH_PART}>/manage"
PROJECTS_EXPLORE_METRIC_QUERY = "m"
PROJECTS_EXPLORE_SAVED_METRIC_QUERY = "sm"

CONNECTIONS_PATH = "/connections"
CONNECTIONS_CREATE_PATH = "/connections/create/"
CONNECTIONS_MANAGE_PATH = f"/connections/<{CONNECTION_ID_PATH_PART}>/manage"

USERS_PATH = "/users"
USERS_HOME_PATH = f"/users/<{USER_PATH_PART}>"

DASHBOARDS_PATH = "/dashboards"
DASHBOARDS_CREATE_PATH = "/dashboards/create"
DASHBOARDS_EDIT_PATH = f"/dashboards/<{DASHBOARD_ID}>/edit"
DASHBOARDS_VIEW_PATH = f"/dashboards/<{DASHBOARD_ID}>/view"

HOME_PATH = "/"

SAVED_METRICS = "/saved_metrics"

SIGN_OUT_URL = "/auth/logout"
UNAUTHORIZED_URL = "/auth/unauthorized"
REDIRECT_TO_LOGIN_URL = "/auth/redirect-to-login"
OAUTH_CODE_URL = "/auth/oauth"


class PathException(Exception):
    pass


def create_path(template: str, **kwargs) -> str:
    for k, v in kwargs.items():
        template = template.replace(f"<{k}>", v)
    return template


def get_path_value(template: str, path: str, key: str) -> str:
    for i, part in enumerate(template.split("/")):
        if part == f"<{key}>":
            path_parts = path.split("/")
            if len(path_parts) >= i:
                return path_parts[i]

    raise PathException(f"{key} is not found in {path} with template {template}")



# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import re
from typing import Iterator, List, Tuple, Type, Union

from sqlalchemy import util
from sqlalchemy.sql import sqltypes
from sqlalchemy.sql.type_api import TypeEngine

SQLType = Union[TypeEngine, Type[TypeEngine]]


class DOUBLE(sqltypes.Float):
    __visit_name__ = "DOUBLE"


class MAP(TypeEngine):
    __visit_name__ = "MAP"

    def __init__(self, key_type: SQLType, value_type: SQLType):
        if isinstance(key_type, type):
            key_type = key_type()
        self.key_type: TypeEngine = key_type

        if isinstance(value_type, type):
            value_type = value_type()
        self.value_type: TypeEngine = value_type

    @property
    def python_type(self):
        return dict


class STRUCT(TypeEngine):
    __visit_name__ = "STRUCT"

    def __init__(self, attr_types: List[Tuple[str, SQLType]]):
        self.attr_types: List[Tuple[str, SQLType]] = []
        for attr_name, attr_type in attr_types:
            if isinstance(attr_type, type):
                attr_type = attr_type()
            self.attr_types.append((attr_name, attr_type))

    @property
    def python_type(self):
        return list


# https://docs.databricks.com/spark/latest/spark-sql/language-manual/sql-ref-datatypes.html
_type_map = {
    "string": sqltypes.VARCHAR,
    "varchar": sqltypes.VARCHAR,
    "char": sqltypes.CHAR,
    # === Boolean ===
    "boolean": sqltypes.BOOLEAN,
    # === Integer ===
    "tinyint": sqltypes.SMALLINT,
    "smallint": sqltypes.SMALLINT,
    "int": sqltypes.INTEGER,
    "integer": sqltypes.INTEGER,
    "bigint": sqltypes.BIGINT,
    "float": sqltypes.FLOAT,
    "double": DOUBLE,
    "decimal": sqltypes.DECIMAL,
    # === Date and time ===
    "date": sqltypes.DATE,
    "timestamp": sqltypes.TIMESTAMP,
    # 'interval year to month':
    # 'interval day to second':
    #
    # === Structural ===
    # 'array': ARRAY,
    "map": MAP,
    "struct": STRUCT
    #
}


def unquote(string: str, quote: str = '"', escape: str = "\\") -> str:
    """
    If string starts and ends with a quote, unquote it
    """
    if string.startswith(quote) and string.endswith(quote):
        string = string[1:-1]
        string = string.replace(f"{escape}{quote}", quote).replace(
            f"{escape}{escape}", escape
        )
    return string


def aware_split(
    string: str,
    delimiter: str = ",",
    maxsplit: int = -1,
    quote: str = '"',
    escaped_quote: str = r"\"",
    open_bracket: str = "<",
    close_bracket: str = ">",
) -> Iterator[str]:
    """
    A split function that is aware of quotes and brackets/parentheses.

    :param string: string to split
    :param delimiter: string defining where to split, usually a comma or space
    :param maxsplit: Maximum number of splits to do. -1 (default) means no limit.
    :param quote: string, either a single or a double quote
    :param escaped_quote: string representing an escaped quote
    :param open_bracket: string, either [, {, < or (
    :param close_bracket: string, either ], }, > or )
    """
    parens = 0
    quotes = False
    i = 0
    if maxsplit < -1:
        raise ValueError(f"maxsplit must be >= -1, got {maxsplit}")
    elif maxsplit == 0:
        yield string
        return
    for j, character in enumerate(string):
        complete = parens == 0 and not quotes
        if complete and character == delimiter:
            if maxsplit != -1:
                maxsplit -= 1
            yield string[i:j]
            i = j + len(delimiter)
            if maxsplit == 0:
                break
        elif character == open_bracket:
            parens += 1
        elif character == close_bracket:
            parens -= 1
        elif character == quote:
            left = j - len(escaped_quote) + 1
            right = j + 1
            if quotes and string[left:right] != escaped_quote:
                quotes = False
            elif not quotes:
                quotes = True
    yield string[i:]


def parse_sqltype(type_str: str) -> TypeEngine:
    type_str = type_str.strip().lower()
    match = re.match(r"^(?P<type>\w+)\s*(?:\<(?P<options>.*)\>)?", type_str)
    if not match:
        util.warn(f"Could not parse type name '{type_str}'")
        return sqltypes.NULLTYPE
    type_name = match.group("type")
    type_opts = match.group("options")

    if type_name == "array":
        item_type = parse_sqltype(type_opts)
        if isinstance(item_type, sqltypes.ARRAY):
            # Multi-dimensions array is normalized in SQLAlchemy, e.g:
            # `ARRAY(ARRAY(INT))` in Spark SQL will become `ARRAY(INT(), dimensions=2)` in SQLAlchemy
            dimensions = (item_type.dimensions or 1) + 1
            return sqltypes.ARRAY(item_type.item_type, dimensions=dimensions)
        return sqltypes.ARRAY(item_type)
    elif type_name == "map":
        key_type_str, value_type_str = aware_split(type_opts)
        key_type = parse_sqltype(key_type_str)
        value_type = parse_sqltype(value_type_str)
        return MAP(key_type, value_type)
    elif type_name == "struct":
        attr_types: List[Tuple[str, SQLType]] = []
        for attr in aware_split(type_opts):
            attr_name, attr_type_str = aware_split(
                attr.strip(), delimiter=":", maxsplit=1
            )
            attr_name = unquote(attr_name)
            attr_type = parse_sqltype(attr_type_str)
            attr_types.append((attr_name, attr_type))
        return STRUCT(attr_types)

    if type_name not in _type_map:
        util.warn(f"Did not recognize type '{type_name}'")
        return sqltypes.NULLTYPE
    type_class = _type_map[type_name]
    type_args = [int(o.strip()) for o in type_opts.split(",")] if type_opts else []
    if type_name in ("time", "timestamp"):
        type_kwargs = dict(timezone=type_str.endswith("with time zone"))
        return type_class(**type_kwargs)
    return type_class(*type_args)



from __future__ import annotations

from typing import Any, Dict, List, Optional

import mitzu.model as M
from dash import html, dcc
from mitzu.helper import value_to_label
import dash_bootstrap_components as dbc
import dash.development.base_component as bc

METRIC_SEGMENTS = "metric_segments"
CHILDREN = "children"
MITZU_LOCATION = "mitzu_location"
WITH_VALUE_CLS = "with_value"

TBL_CLS = "small text mh-0 align-middle"
TBL_CLS_WARNING = "small text-danger mh-0 fw-bold align-middle"
TBL_HEADER_CLS = "small mh-0 text-nowrap fw-bold align-middle"


MISSING_RESOURCE_CSS = "border border-2 border-warning"


def get_enums(path: str, discovered_project: M.DiscoveredProject) -> List[Any]:
    event_field_def = find_event_field_def(path, discovered_project)
    if event_field_def is not None:
        res = event_field_def._enums
        return res if res is not None else []
    return []


def find_event_field_def(
    path: str, discovered_project: M.DiscoveredProject
) -> M.EventFieldDef:
    path_parts = path.split(".")
    event_name = path_parts[0]
    event_def = discovered_project.get_event_def(event_name)
    field_name = ".".join(path_parts[1:])

    for field, event_field_def in event_def._fields.items():
        if field._get_name() == field_name:
            return event_field_def
    raise Exception(f"Invalid property path: {path}")


def get_event_names(segment: Optional[M.Segment]) -> List[str]:
    if segment is None:
        return []
    if isinstance(segment, M.SimpleSegment):
        if segment._left is None:
            return []
        return [segment._left._event_name]
    elif isinstance(segment, M.ComplexSegment):
        return get_event_names(segment._left) + get_event_names(segment._right)
    else:
        raise Exception(f"Unsupported Segment Type: {type(segment)}")


def get_property_name_label(field_name: str, max_length: int = 100) -> str:
    field_name = field_name
    res = field_name.replace("_", " ").replace(".", " > ").title()
    if len(res) > max_length:
        res = f"... {res[-max_length:]}"
    return res


def get_final_all_inputs(
    all_inputs: Dict[str, Any], ctx_input_list: List[Dict]
) -> Dict[str, Any]:
    res: Dict[str, Any] = all_inputs
    res[METRIC_SEGMENTS] = {}
    for ipt in ctx_input_list:
        if type(ipt) == list:
            for sub_input in ipt:
                if type(sub_input) == dict:
                    sub_input_id = sub_input["id"]
                    index = sub_input_id["index"]
                    input_type = sub_input_id["type"]
                    curr = res[METRIC_SEGMENTS]
                    for sub_index in index.split("-"):
                        sub_index = int(sub_index)
                        if CHILDREN not in curr:
                            curr[CHILDREN] = {}
                        curr = curr[CHILDREN]
                        if sub_index not in curr:
                            curr[sub_index] = {}
                        curr = curr[sub_index]
                    curr[input_type] = sub_input["value"]
                else:
                    raise ValueError(f"Invalid sub-input type: {type(sub_input)}")

    return res


def create_form_property_input(
    property: str,
    index_type: str,
    icon_cls: Optional[str] = None,
    component_type: bc.Component = dbc.Input,
    input_lg: int = 3,
    input_sm: int = 12,
    label_lg: int = 3,
    label_sm: int = 12,
    justify: Optional[str] = None,
    read_only: bool = False,
    **kwargs,
):
    if "size" not in kwargs and component_type not in [dbc.Checkbox, dcc.Dropdown]:
        kwargs["size"] = "sm"
    if (
        component_type in [dbc.Input, dbc.Textarea]
        and kwargs.get("placeholder") is None
    ):
        kwargs["placeholder"] = value_to_label(property)

    if icon_cls is not None:
        label_children = [
            html.B(className=f"{icon_cls} me-1"),
            value_to_label(property),
            "*" if kwargs.get("required") and not read_only else "",
        ]
    else:
        label_children = [value_to_label(property)]
    if read_only:
        component_type = dbc.Input
        kwargs["readonly"] = True
        kwargs["disabled"] = True
        if "data" in kwargs.keys():
            del kwargs["data"]
    else:
        kwargs["id"] = {"type": index_type, "index": property}
    return dbc.Row(
        [
            dbc.Label(label_children, class_name="fw-bold", sm=label_sm, lg=label_lg),
            dbc.Col(
                component_type(
                    **kwargs,
                ),
                sm=input_sm,
                lg=input_lg,
            ),
        ],
        class_name="mb-3",
        justify=justify,
    )



from dash import register_page
import dash_bootstrap_components as dbc
from dash import html
import dash.development.base_component as bc
import flask
import mitzu.webapp.dependencies as DEPS
import mitzu.webapp.navbar as NB
from mitzu.webapp.auth.decorator import restricted_layout
import mitzu.webapp.pages.paths as P
from typing import List, cast
import traceback


PROJECTS_CONTAINER = "projects_container"
PROJECTS_ROW = "projects_row"
PROJECT_CARD_TITLE = "project_card_title"

register_page(
    __name__,
    path=P.PROJECTS_PATH,
    title="Mitzu - Explore",
)


@restricted_layout
def layout(**query_params) -> bc.Component:
    projects = create_projects_children()

    return html.Div(
        [
            NB.create_mitzu_navbar("explore-navbar", []),
            dbc.Container(
                children=[
                    dbc.Row(
                        [
                            dbc.Col(
                                html.H4(
                                    "Select a project for exploration",
                                    className="card-title",
                                ),
                                width="auto",
                            ),
                            dbc.Col(
                                dbc.Button(
                                    children=[
                                        html.I(className="bi bi-plus-circle me-1"),
                                        "Add project",
                                    ],
                                    href=P.PROJECTS_CREATE_PATH,
                                ),
                                width="auto",
                                class_name="ms-auto",
                            ),
                        ]
                    ),
                    html.Hr(),
                    html.Div(children=projects, id=PROJECTS_CONTAINER),
                    html.Hr(),
                ]
            ),
        ]
    )


def create_projects_children() -> List[bc.Component]:
    depenednecies = cast(
        DEPS.Dependencies, flask.current_app.config.get(DEPS.CONFIG_KEY)
    )
    project_ids = depenednecies.storage.list_projects()

    projects = []
    if len(project_ids) > 0:
        for p in project_ids:
            try:
                projects.append(create_project_selector(p, depenednecies))
            except Exception as exc:
                traceback.print_exc()
                projects.append(
                    dbc.Col(
                        dbc.Card(
                            dbc.CardBody(html.P(str(exc), className="text-danger")),
                            class_name="mb-3",
                        ),
                        lg=3,
                        sm=12,
                    )
                )

        return dbc.Row(children=projects, id=PROJECTS_ROW)

    return html.H4(
        "You don't have any projects yet...", className="card-title text-center"
    )


def create_project_selector(project_id: str, deps: DEPS.Dependencies) -> bc.Component:
    project = deps.storage.get_project(project_id)
    discovered_project = project._discovered_project.get_value()

    tables = len(project.event_data_tables)
    events = len(discovered_project.get_all_events()) if discovered_project else 0
    project_jumbotron = dbc.Col(
        dbc.Card(
            dbc.CardBody(
                [
                    html.H4(
                        project.project_name,
                        className="card-title",
                        id=PROJECT_CARD_TITLE,
                    ),
                    html.Hr(),
                    html.Img(
                        src=f"/assets/warehouse/{str(project.connection.connection_type.name).lower()}.png",
                        height=40,
                    ),
                    html.P(f"This project has {events} events in {tables} datasets."),
                    html.P(project.description),
                    html.Div(
                        [
                            dbc.Button(
                                "Explore",
                                color="primary",
                                class_name="me-3",
                                href=P.create_path(
                                    P.PROJECTS_EXPLORE_PATH, project_id=project_id
                                ),
                            ),
                            dbc.Button(
                                "Manage",
                                color="secondary",
                                href=P.create_path(
                                    P.PROJECTS_MANAGE_PATH, project_id=project_id
                                ),
                            ),
                        ],
                    ),
                ]
            ),
            class_name="mb-3",
        ),
        lg=3,
        sm=12,
    )
    return project_jumbotron



from __future__ import annotations

import logging
import os
import sys
from datetime import datetime
from typing import Any, Optional, Dict
import inspect

import mitzu.model as M
from uuid import uuid4

LOGGER = logging.getLogger(name="mitzu_logger")
LOGGER.addHandler(logging.StreamHandler(sys.stdout))
LOGGER.setLevel(os.getenv("LOG_LEVEL", logging.INFO))


def value_to_label(value: str) -> str:
    return value.title().replace("_", " ")


def find_notebook_globals(hint_string: str) -> Optional[Dict]:
    for stk in inspect.stack():
        glbs = stk[0].f_globals

        if (hint_string in glbs) and "find_notebook_globals" not in glbs:
            LOGGER.debug("Found globals on stack")
            return glbs
    LOGGER.warning("Globals not found on stack.")
    return None


def parse_datetime_input(val: Any, def_val: Optional[datetime]) -> Optional[datetime]:
    if val is None:
        return def_val
    if type(val) == str:
        return datetime.fromisoformat(val)
    elif type(val) == datetime:
        return val
    else:
        raise ValueError(f"Invalid argument type for datetime parse: {type(val)}")


def get_segment_project(segment: M.Segment) -> M.Project:
    if isinstance(segment, M.SimpleSegment):
        left = segment._left
        if isinstance(left, M.EventDef):
            return left._project
        elif isinstance(left, M.EventFieldDef):
            return left._project
        else:
            raise ValueError(f"Segment's left value is of invalid type: {type(left)}")
    elif isinstance(segment, M.ComplexSegment):
        return get_segment_project(segment._left)
    else:
        raise ValueError(f"Segment is of invalid type: {type(segment)}")


def create_unique_id():
    return str(uuid4())[-12:]



from __future__ import annotations

from dataclasses import dataclass
import pandas as pd
from typing import Any, Callable, Optional
import mitzu.model as M


TTC_RANGE_1_SEC = 600
TTC_RANGE_2_SEC = 7200
TTC_RANGE_3_SEC = 48 * 3600


X_AXIS_COL = "x"
Y_AXIS_COL = "y"
TEXT_COL = "_text"
COLOR_COL = "_color"
TOOLTIP_COL = "_tooltip"


def retention_period_label(val: int, metric: M.Metric) -> str:
    if isinstance(metric, M.RetentionMetric):
        return f"{val} to {val+ metric._retention_window.value} {metric._retention_window.period.name.lower()}"
    return str(val)


def fix_date_label(val: pd.Timestamp, metric: M.Metric) -> str:
    if metric._time_group not in (
        M.TimeGroup.SECOND,
        M.TimeGroup.MINUTE,
        M.TimeGroup.HOUR,
    ):
        res = str(val.date())
        if metric._chart_type == M.SimpleChartType.HEATMAP:
            res = res + "."
        return res
    return val


@dataclass(frozen=True)
class SimpleChart:

    title: str
    x_axis_label: str
    y_axis_label: str
    color_label: str
    yaxis_ticksuffix: str
    hover_mode: str
    chart_type: M.SimpleChartType
    dataframe: pd.DataFrame
    x_axis_labels_func: Optional[Callable[[Any, M.Metric], Any]] = None
    y_axis_labels_func: Optional[Callable[[Any, M.Metric], Any]] = None
    color_labels_func: Optional[Callable[[Any, M.Metric], Any]] = None



from __future__ import annotations

from typing import Any, List, cast

import mitzu.adapters.generic_adapter as GA
import mitzu.adapters.sqlalchemy.databricks.sqlalchemy.datatype as DA_T
import mitzu.model as M
from mitzu.adapters.sqlalchemy.databricks import sqlalchemy  # noqa: F401
from mitzu.adapters.sqlalchemy_adapter import SQLAlchemyAdapter, FieldReference
from mitzu.adapters.helper import pdf_string_json_array_to_array
from mitzu.helper import LOGGER

import sqlalchemy as SA
import sqlalchemy.sql.expression as EXP
import sqlalchemy.sql.sqltypes as SA_T
from sqlalchemy.sql.type_api import TypeEngine


class DatabricksAdapter(SQLAlchemyAdapter):
    def __init__(self, project: M.Project):
        super().__init__(project)

    def get_engine(self) -> Any:
        con = self.project.connection
        if self._engine is None:
            if con.url is None:
                url = self._get_connection_url(con)
            else:
                url = con.url
            http_path = con.extra_configs.get("http_path")
            if http_path is None:
                raise Exception(
                    "Connection extra_configs must contain http_path. (extra_configs={'http_path':'<path>'}"
                )
            self._engine = SA.create_engine(url, connect_args={"http_path": http_path})
        return self._engine

    def map_type(self, sa_type: Any) -> M.DataType:
        if isinstance(sa_type, DA_T.MAP):
            return M.DataType.MAP
        if isinstance(sa_type, DA_T.STRUCT):
            return M.DataType.STRUCT
        return super().map_type(sa_type)

    def _parse_map_type(
        self, sa_type: Any, name: str, event_data_table: M.EventDataTable
    ) -> M.Field:
        if event_data_table.discovery_settings is None:
            raise ValueError("Missing discovery settings")

        LOGGER.debug(f"Discovering map: {name}")
        map: DA_T.MAP = cast(DA_T.MAP, sa_type)
        if map.value_type in (DA_T.STRUCT, DA_T.MAP):
            raise Exception(
                f"Compounded map types are not supported: map<{map.key_type}, {map.value_type}>"
            )
        cte = self._get_dataset_discovery_cte(event_data_table)
        F = SA.func
        map_keys_func = F.array_distinct(
            F.flatten(F.collect_set(F.map_keys(cte.columns[name])))
        )

        max_cardinality = event_data_table.discovery_settings.max_map_key_cardinality
        q = SA.select(
            columns=[
                SA.case(
                    [(F.size(map_keys_func) < max_cardinality, map_keys_func)],
                    else_=None,
                ).label("sub_fields")
            ]
        )
        df = self.execute_query(q)
        if df.shape[0] == 0:
            # The map typa has no values
            return M.Field(_name=name, _type=M.DataType.MAP)
        if df.iat[0, 0] is None:
            # The total amount of map keys are more than the allowed from the EDT
            keys = []
        else:
            keys = df.iat[0, 0].tolist()

        sf_type = self.map_type(map.value_type)
        sub_fields: List[M.Field] = [M.Field(key, sf_type) for key in keys]
        return M.Field(_name=name, _type=M.DataType.MAP, _sub_fields=tuple(sub_fields))

    def _get_struct_type(self) -> TypeEngine:
        return DA_T.STRUCT

    def _get_conv_aggregation(
        self, metric: M.Metric, cte: EXP.CTE, first_cte: EXP.CTE
    ) -> Any:
        if metric._agg_type == M.AggType.PERCENTILE_TIME_TO_CONV:
            if metric._agg_param is None or 0 < metric._agg_param > 100:
                raise ValueError(
                    "Conversion percentile parameter must be between 0 and 100"
                )
            t1 = first_cte.columns.get(GA.CTE_DATETIME_COL)
            t2 = cte.columns.get(GA.CTE_DATETIME_COL)
            return SA.func.percentile_approx(
                SA.cast(t2 - t1, SA_T.BIGINT), metric._agg_param / 100.0, 100
            )
        if metric._agg_type == M.AggType.AVERAGE_TIME_TO_CONV:
            t1 = first_cte.columns.get(GA.CTE_DATETIME_COL)
            t2 = cte.columns.get(GA.CTE_DATETIME_COL)
            return SA.func.avg(SA.cast(t2 - t1, SA_T.BIGINT))
        else:
            return super()._get_conv_aggregation(metric, cte, first_cte)

    def _get_dynamic_datetime_interval(
        self,
        field_ref: FieldReference,
        value_field_ref: FieldReference,
        time_group: M.TimeGroup,
    ) -> Any:
        if time_group == M.TimeGroup.SECOND:
            return SA.func.from_unixtime(
                SA.func.unix_timestamp(field_ref) + value_field_ref
            )
        if time_group == M.TimeGroup.MINUTE:
            return SA.func.from_unixtime(
                SA.func.unix_timestamp(field_ref) + value_field_ref * 60
            )
        if time_group == M.TimeGroup.HOUR:
            SA.func.from_unixtime(
                SA.func.unix_timestamp(field_ref) + value_field_ref * 3600
            )
        if time_group == M.TimeGroup.DAY:
            return SA.func.date_add(field_ref, value_field_ref)
        if time_group == M.TimeGroup.WEEK:
            return SA.func.date_add(field_ref, value_field_ref * 7)
        if time_group == M.TimeGroup.MONTH:
            return SA.func.add_months(field_ref, value_field_ref)
        if time_group == M.TimeGroup.QUARTER:
            return SA.func.add_months(field_ref, value_field_ref * 4)
        if time_group == M.TimeGroup.YEAR:
            return SA.func.add_months(field_ref, value_field_ref * 12)

        raise ValueError(
            f"{time_group} is not supported for databricks dynamic time intervals"
        )

    def _get_column_values_df(
        self,
        event_data_table: M.EventDataTable,
        fields: List[M.Field],
        event_specific: bool,
    ):
        df = super()._get_column_values_df(event_data_table, fields, event_specific)
        return pdf_string_json_array_to_array(df)

    def _get_distinct_array_agg_func(self, field_ref: FieldReference) -> Any:
        return SA.func.to_json(SA.func.array_agg(SA.distinct(field_ref)))



from __future__ import annotations

from datetime import datetime
from typing import Any, List, Union, cast

import mitzu.adapters.generic_adapter as GA
import mitzu.model as M
import pandas as pd
import trino.sqlalchemy.datatype as SA_T
from mitzu.adapters.helper import (
    dataframe_str_to_datetime,
    pdf_string_json_array_to_array,
)
from mitzu.adapters.sqlalchemy_adapter import FieldReference, SQLAlchemyAdapter
from mitzu.helper import LOGGER
from sqlalchemy.sql.type_api import TypeEngine
import sqlalchemy as SA
import sqlalchemy.sql.expression as EXP


class TrinoAdapter(SQLAlchemyAdapter):
    def __init__(self, project: M.Project):
        super().__init__(project)

    def get_conversion_df(self, metric: M.ConversionMetric) -> pd.DataFrame:
        df = super().get_conversion_df(metric)
        df = dataframe_str_to_datetime(df, GA.DATETIME_COL)
        for index in range(1, len(metric._conversion._segments) + 1):
            df[f"{GA.AGG_VALUE_COL}_{index}"] = df[
                f"{GA.AGG_VALUE_COL}_{index}"
            ].astype(float)
        return df

    def get_segmentation_df(self, metric: M.SegmentationMetric) -> pd.DataFrame:
        df = super().get_segmentation_df(metric)
        df = dataframe_str_to_datetime(df, GA.DATETIME_COL)
        df[GA.AGG_VALUE_COL] = df[GA.AGG_VALUE_COL].astype(float)
        return df

    def get_retention_df(self, metric: M.RetentionMetric) -> pd.DataFrame:
        df = super().get_retention_df(metric)
        df = dataframe_str_to_datetime(df, GA.GROUP_COL)
        df = dataframe_str_to_datetime(df, GA.DATETIME_COL)
        df[GA.AGG_VALUE_COL] = df[GA.AGG_VALUE_COL].astype(float)
        return df

    def execute_query(self, query: Any) -> pd.DataFrame:
        if type(query) != str:
            query = str(query.compile(compile_kwargs={"literal_binds": True}))
        return super().execute_query(query=query)

    def get_field_reference(
        self,
        field: M.Field,
        event_data_table: M.EventDataTable = None,
        sa_table: Union[SA.Table, EXP.CTE] = None,
    ) -> FieldReference:
        if sa_table is None and event_data_table is not None:
            sa_table = self.get_table(event_data_table)
        if sa_table is None:
            raise ValueError("Either sa_table or event_data_table has to be provided")

        if field._parent is not None and field._parent._type == M.DataType.MAP:
            map_key = field._name
            property = SA.literal_column(f"{sa_table.name}.{field._parent._get_name()}")
            return SA.func.element_at(property, map_key)

        return super().get_field_reference(field, event_data_table, sa_table)

    def map_type(self, sa_type: Any) -> M.DataType:
        if isinstance(sa_type, SA_T.ROW):
            return M.DataType.STRUCT
        if isinstance(sa_type, SA_T.MAP):
            return M.DataType.MAP
        return super().map_type(sa_type)

    def _parse_map_type(
        self,
        sa_type: Any,
        name: str,
        event_data_table: M.EventDataTable,
    ) -> M.Field:
        if event_data_table.discovery_settings is None:
            raise ValueError("Missing discovery settings")

        LOGGER.debug(f"Discovering map: {name}")
        map: SA_T.MAP = cast(SA_T.MAP, sa_type)
        if map.value_type in (SA_T.ROW, SA_T.MAP):
            raise Exception(
                f"Compounded map types are not supported: map<{map.key_type}, {map.value_type}>"
            )
        cte = self._get_dataset_discovery_cte(event_data_table)
        F = SA.func
        map_keys_func = F.array_distinct(
            F.flatten(F.array_agg(F.distinct(F.map_keys(cte.columns[name]))))
        )

        max_cardinality = event_data_table.discovery_settings.max_map_key_cardinality
        q = SA.select(
            columns=[
                SA.case(
                    [(F.cardinality(map_keys_func) < max_cardinality, map_keys_func)],
                    else_=None,
                ).label("sub_fields")
            ]
        )
        df = self.execute_query(q)
        if df.shape[0] == 0:
            return M.Field(_name=name, _type=M.DataType.MAP)
        keys = df.iat[0, 0]
        sf_type = self.map_type(map.value_type)
        sub_fields: List[M.Field] = [M.Field(key, sf_type) for key in keys]
        return M.Field(_name=name, _type=M.DataType.MAP, _sub_fields=tuple(sub_fields))

    def _generate_time_series_column(self, dt: datetime) -> Any:
        return SA.literal_column(f"timestamp '{dt}'")

    def _get_struct_type(self) -> TypeEngine:
        return SA_T.ROW

    def _get_distinct_array_agg_func(self, field_ref: FieldReference) -> Any:
        return SA.func.array_agg(SA.distinct(field_ref))

    def _get_column_values_df(
        self,
        event_data_table: M.EventDataTable,
        fields: List[M.Field],
        event_specific: bool,
    ) -> pd.DataFrame:
        df = super()._get_column_values_df(
            event_data_table=event_data_table,
            fields=fields,
            event_specific=event_specific,
        )
        return pdf_string_json_array_to_array(df)

    def _correct_timestamp(self, dt: datetime) -> Any:
        return SA.text(f"timestamp '{dt}'")

    def _get_datetime_interval(
        self, field_ref: FieldReference, timewindow: M.TimeWindow
    ) -> Any:
        return SA.func.date_add(
            timewindow.period.name.lower(),
            timewindow.value,
            field_ref,
        )

    def _get_dynamic_datetime_interval(
        self,
        field_ref: FieldReference,
        value_field_ref: FieldReference,
        time_group: M.TimeGroup,
    ) -> Any:
        return SA.func.date_add(time_group.name.lower(), value_field_ref, field_ref)

    def _get_conv_aggregation(
        self, metric: M.Metric, cte: EXP.CTE, first_cte: EXP.CTE
    ) -> Any:
        if metric._agg_type == M.AggType.PERCENTILE_TIME_TO_CONV:
            if metric._agg_param is None or 0 < metric._agg_param > 100:
                raise ValueError(
                    "Conversion percentile parameter must be between 0 and 100"
                )
            t1 = first_cte.columns.get(GA.CTE_DATETIME_COL)
            t2 = cte.columns.get(GA.CTE_DATETIME_COL)
            return SA.func.approx_percentile(
                SA.func.date_diff("second", t1, t2), metric._agg_param / 100.0
            )
        if metric._agg_type == M.AggType.AVERAGE_TIME_TO_CONV:
            t1 = first_cte.columns.get(GA.CTE_DATETIME_COL)
            t2 = cte.columns.get(GA.CTE_DATETIME_COL)
            return SA.func.avg(SA.func.date_diff("second", t1, t2))
        else:
            return super()._get_conv_aggregation(metric, cte, first_cte)
